###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 2000-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## Portions Copyright (C) 1998-2001 slennox <slennox@egghelp.org>
## Originally based on and inspired by botnetop.tcl v2.10 and v2.23.
## Original botnetop.tcl can be found at http://www.egghelp.org/
##
## First cleanup and conversion of botnetop.tcl v2.10 to a wolfpack
## module (04 July 2000).
## Complete rewrite and merge of v2.23 changes (06 Jan 2005).
##
## $Id: botnetop.tcl,v 1.37 2005/01/16 11:41:30 tothwolf Exp $
##

# name: botnetop
# version: 2.23.3
# author: Tothwolf <tothwolf@concentric.net>
# config: botnetop.conf
# description: Allows bots to op each other via the botnet
# requires: wpLog killUtimerCommand

##
## NOTE: Contact Tothwolf if you find any bugs. This is a redesigned
##       botnetop script and is very different from slennox's version.
##

#
# Notes from slennox's botnetop.tcl:
#
# This script was written mainly as an alternative to getops.tcl. It allows
# linked bots to securely give ops to one another on their channels. It
# also automatically invites linked bots to channels which are invite-only,
# shares keys with linked bots for channels which are keyed, raises a
# channel's limit if a bot can't get in because the channel is full, unbans
# a bot if it can't get in because it matches a ban, and automatically adds
# new hostmasks when a bot's user@host info has changed. There is a delayed
# op mechanism to prevent mode +o flooding whenever a bot joins a channel,
# and op/invite checking routines are significantly more secure than those
# in getops.tcl.
#
# botnetop.tcl v2.xx is compatible with bots running older versions of
# botnetop.tcl (v1.00 and up) and bots running the botnetop.tcl component
# of netbots.tcl.
#
# Note that this script was developed on the eggdrop 1.3, 1.4, and 1.6
# series and may not work properly on other versions.
#
# v2.00 - New standalone release. Contains refinements and features from
#         the netbots.tcl version of botnetop.tcl.
# v2.02 - Improved support for leading +-^= characters in hostmasks.
#         Fixed 'islinked' problem on eggdrop 1.3.24 and earlier.
# v2.10 - Requests to unopped bots were causing high CPU usage and botnet
#         traffic on some large opless botnets - rewrote proc to request
#         from opped bots only.
#         Added bop_modeop feature to provide faster opping.
#         Convinced myself to change putquick back to pushmode (suits
#         bop_modeop better).
# v2.20 - pushmode sanity checking in eggdrop 1.4+ prevents bop_osync from
#         working properly - putquick is now automatically used when
#         bop_osync is enabled on a 1.4+ bot.
#         Fixed bug in bop_reqop that was causing "bot is not in the
#         botnet" errors.
#         Updated to support eggdrop 1.6 (includes 'bind need' support),
#         requires only a single need bind for all requests.
#         Variables are now cleared after removing a channel.
#         Removed unnecessary botonchan checks throughout components where
#         botisop already checks for that.
#         Removed all unnecessary use of parentheses.
#         bop_needi was not defined as a global variable in bop_settimer.
# v2.21 - Universal need bind could cause bop_ errors on 1.6 bots.
# v2.23 - Added bop_clearneeds to eliminate need-related errors.
#         Made bop_settimer aggressive again.
#         Removed unused putquick compatibility proc.
#
# About the delay and maxreq settings: botnetop.tcl uses two
# methods to reduce/eliminate mode +o flooding - delay is used by the
# bot giving ops, while maxreq is used by the bot asking for ops. You
# can use both features, one or the other, or disable both. Below is an
# explanation of some possible combinations.
#
# delay 20; maxreq 0: all bots will ask a non-opped bot if it needs
# ops after a delay of between 0-20 seconds (different bots will delay for
# different amounts of time). The bot that needs ops will respond to all
# offers.
#
# delay 0; maxreq 2: all bots will ask a non-opped bot if it needs
# ops without any delay. The bot that needs ops will respond to a maximum
# of 2 offers.
#
# delay 20; maxreq 3: all bots will ask a non-opped bot if it needs
# ops after a delay of between 0-20 seconds. The bot that needs ops will
# respond to a maximum of 3 offers.
#
# delay 0; maxreq 0: all bots will ask a non-opped bot if it needs
# ops without any delay. The bot that needs ops will respond to all offers.


# FIXME: some ircds allow invite to bypass +l, +k and +b
#
#        $configData(tryinvite)
#
# FIXME: make logging log for channel?
#
# FIXME: support 1.6.0+ 'bind need'
# if {$numversion >= 1060000} then {
#   bind need -|- * botneeds
# }
#
# FIXME: merge code from 'requestops' and 'letmein' ?
#
# FIXME: +l add additional amount?
#        and lower once bot joins?
#        and kick someone who isn't supposed to join? ;)


##
## Check if an array contains some data
##   Used for 'Need'; check if a need-* setting has already been processed
##   Used for 'Join'; check if we should respond to join notifies
##
## Args: arrayName, action {set|clear|check}, what {key|level|invite|unban},
##       channel
## Returns: 0, 1, -1
##
proc checkActionData {arrayName action what chan} {
  upvar $arrayName array

  set chan [string tolower $chan]

  switch -exact -- $action {
    set {
      set array($chan,$what) 1
      return 1
    }
    clear {
      if {[info exists array($chan,$what)]} then {
        unset array($chan,$what)
        return 1
      }
      return -1
    }
    check {
      if {[info exists array($chan,$what)]} then {
        return 1
      }
    }
  }
  return 0
}

##
## Called by the bot's need-* options
##
## Args:    channel, need
## Returns: nothing
##
proc botneeds {chan need} {
  if {![string compare op $need]} then {
    foreach bot [chanlist $chan b] {
      if {([isop $bot $chan]) &&
          ([matchattr [set hand [nick2hand $bot $chan]] o|o $chan]) &&
          ([islinked $hand])} then {
        putbot $hand "reqops $chan"
      }
    }
  } else {
    letmein $chan $need
  }
  return
}

##
## Called by the bot's need-* options when the bot needs in a channel
##
## Args:    channel, need
## Returns: nothing
##
proc letmein {chan need} {
  global botnick botname
  variable configData
  variable Need

  set bots [bots]
  if {([string compare "" $bots]) && (![botonchan $chan]) &&
      ([regexp {(key|invite|limit|unban)} $need]) &&
      (![checkActionData Need check $need $chan])} then {

    switch -exact -- $need {
      key {
        set what "wantkey $chan $botnick"
      }
      invite {
        if {[string compare "" $botname]} then {
          set what "wantinvite $chan $botnick [string trimleft [lindex [split $botname !] 1] "~+-^="]"
        } else {
          set what "wantinvite $chan $botnick"
        }
      }
      limit {
        set what "wantlimit $chan $botnick"
      }
      unban {
        set what "wantunban $chan $botnick $botname"
      }
    }

    set list ""
    foreach bot $bots {
      if {([matchattr $bot b]) && ([matchattr $bot o|o $chan])} then {
        putbot $bot $what
        lappend list $bot
      }
    }

    if {[string compare "" $list]} then {
      checkActionData Need set $need $chan
      utimer 30 [namespace code [list checkActionData Need clear $need $chan]]

      if {$configData(log) >= 1} then {
        regsub -all -- { } [join $list] {, } list

        switch -exact -- $need {
          key {
            wpLog o * [mc "requested key for %s from %s" $chan $list]
          }
          invite {
            wpLog o * [mc "requested invite to %s from %s" $chan $list]
          }
          limit {
            wpLog o * [mc "requested limit raise on %s from %s" $chan $list]
          }
          unban {
            wpLog o * [mc "requested unban on %s from %s" $chan $list]
          }
        }
      }
    }
  }
  return
}

##
## Request ops from a bot that gets op'd
##
## Args:    nick, user@host, handle, channel, mode, target
## Returns: 0
##
proc mode:watchmodes {nick uhost hand chan mode {target ""}} {
  global botname
  variable configData
  variable WatchModes

  if {![string compare "" $target]} then {
    set target [lindex $mode 1]
    set mode [lindex $mode 0]
  }

  switch -exact -- $mode {
    +o {
      # If we see a bot get ops and we need ops, ask it first
      if {![isbotnick $target]} then {
        if {![botisop $chan]} then {
          set targetHand [nick2hand $target $chan]
          if {([validuser $targetHand]) &&
              ([matchattr $targetHand b]) &&
              ([matchattr $targetHand o|o $chan]) &&
              ([islinked $targetHand])} then {
            putbot $targetHand "reqops $chan"
          }
        }
      } else {

# FIXME: remember bot asking if we want ops and
#        if it doesn't have a matching hostmask add it here when it ops us?

# FIXME: remember op/need-* requests when not op'd and respond to them
#        after gaining ops

        # Bot got ops, check for and remove any bans matching the bot
        if {$configData(checkbans)} then {
          foreach chanban [chanbans $chan] {
            foreach {ban from time} $chanban {break}
            if {[string match [string tolower $ban] \
                 [string tolower $botname]]} then {
              pushmode $chan -b $ban
            }
          }
        }
      }
    }
    -o {
      # If we get deoped, ask for ops immediately
# FIXME: does Eggdrop now do this on its own? This may be redundant.
      if {([isbotnick $target]) && ([botonchan $chan]) &&
          (![botisop $chan])} then {
        botneeds $chan op
      }
    }
    +k {
      # Some braindead ircds hide the channel key as '*' in the mode change.
      # This will cause eggdrop to refresh its cached channel mode data.
# FIXME: could this cause the bot to flood itself?
      if {($configData(fixkey)) && (![string compare "*" $target])} then {
        putserv "MODE $chan"
      }
    }
    +l {
      # Tell a bot that its ok to join the channel now
      if {[isbotnick $nick]} then {
        set chanLower [string tolower $chan]
        set names [array names WatchModes "$chanLower,limit,*"]
        if {[string compare "" $names]} then {
          foreach name $names {
            putbot $WatchModes($name) "donelimit $chan"
          }
          array unset WatchModes "$chanLower,limit,*"
        }
      }
    }
    -b {
      # Tell a bot that its ok to join the channel now
      if {[isbotnick $nick]} then {
        set chanLower [string tolower $chan]
        set names [array names WatchModes "$chanLower,unban,*"]
        if {[string compare "" $names]} then {
          foreach name $names {
            putbot $WatchModes($name) "doneunban $chan"
          }
          array unset WatchModes "$chanLower,unban,*"
        }
      }
    }
  }
  return 0
}

##
## Request ops from bots when this bot links to the botnet
##
## Args:    bot, via
## Returns: 0
##
proc link:checkneeds {bot via} {
  global botnet-nick

# FIXME: add support for need-key/invite/limit here

  if {(![string compare $bot ${botnet-nick}]) ||
      (![string compare $via ${botnet-nick}]) &&
      (![regexp -- {.*RequestOpsChannels.*} [utimers]])} then {
    utimer 2 [namespace code RequestOpsChannels]
  }
  return 0
}

##
## Check all channels to see if the bot needs ops
##
## Args:    none
## Returns: nothing
##
proc RequestOpsChannels {} {
  foreach chan [channels] {

    CheckNeed $chan

    if {([botonchan $chan]) && (![botisop $chan])} then {
      botneeds $chan op
    }
  }
  return
}

##
## Ask a bot on join if they need ops, or set defaults
##
## Args:    nick, user@host, handle, channel
## Returns: 0
##
proc join:askops {nick uhost hand chan} {
  variable configData
  variable AskTimer
  variable ChanKey

  if {![isbotnick $nick]} then {
    if {([matchattr $hand b]) && ([matchattr $hand o|o $chan]) &&
        (![matchattr $hand d|d $chan])} then {
      set chanLower [string tolower $chan]
      if {![info exists AskTimer($hand,$chanLower)]} then {
        set AskTimer($hand,$chanLower) 1
        if {(!$configData(delay)) || ([LowBots $chan])} then {
          utimer 5 [namespace code [list AskBot $hand $chan]]
        } else {
          utimer [expr [rand $configData(delay)] + 5] [namespace code [list AskBot $hand $chan]]
        }
      }
    }
  } else {
    set chanLower [string tolower $chan]
    if {[info exists ChanKey($chanLower)]} then {
      unset ChanKey($chanLower)
    }
    CheckNeed $chan
  }
  return 0
}

##
## Botnet reqops
##
## Args:    bot name, command, arg
## Returns: 0
##
proc bot:reqops {bot cmd arg} {
  variable configData
  variable AskTimer

  set chan [lindex [split $arg] 0]
  if {([validchan $chan]) && ([matchattr $bot b]) &&
      ([matchattr $bot o|o $chan]) && (![matchattr $bot d|d $chan])} then {
    set chanLower [string tolower $chan]
    if {![info exists AskTimer($bot,$chanLower)]} then {
      set AskTimer($bot,$chanLower) 1
      if {(!$configData(delay)) || ([LowBots $chan])} then {
        utimer 2 [namespace code [list AskBot $bot $chan]]
      } else {
        utimer [expr [rand $configData(delay)] + 2] [namespace code [list AskBot $bot $chan]]
      }
    }
  }
  return 0
}

##
## Check the number of op'd bots in the given channel
##
## Args:    channel
## Returns: 1 if the value set with 'lowbots' or more bots are op'd
##            in the given channel
##          0 otherwise
##
proc LowBots {chan} {
  variable configData

  if {$configData(lowbots) > 0} then {
    set bots 0
    foreach bot [chanlist $chan b] {
      if {(![isbotnick $bot]) && ([isop $bot $chan])} then {
        incr bots
        # Check this here too in case we have a lot of bots on the channel
        if {$bots >= $configData(lowbots)} then {
          return 0
        }
      }
    }
    if {$bots < $configData(lowbots)} then {
      return 1
    }
  }
  return 0
}

##
## AskBot
##
## Args:    handle, channel
## Returns: nothing
##
proc AskBot {hand chan} {
  global botnick
  variable AskTimer

  set chanLower [string tolower $chan]
  if {[info exists AskTimer($hand,$chanLower)]} then {
    unset AskTimer($hand,$chanLower)
  }
  if {([islinked $hand]) && ([validchan $chan]) && ([botonchan $chan]) &&
      ([botisop $chan]) && ([matchattr $hand b]) &&
      ([matchattr $hand o|o $chan]) && (![matchattr $hand d|d $chan])} then {
    putbot $hand "doyawantops $chan $botnick"
  }
  return
}

##
## Answer a request if this bot wants ops
##
## Args:    bot name, command, arg
## Returns: 0
##
proc bot:doyawantops {bot cmd arg} {
  global botnick botname
  variable configData
  variable OpRequests

  set arg [split $arg]
  set chan [lindex $arg 0]
  set fromnick [lindex $arg 1]
  if {([validchan $chan]) && ([botonchan $chan]) &&
      (![botisop $chan])} then {
    set chanLower [string tolower $chan]
    if {((($configData(maxreq)) &&
          ($OpRequests($chanLower) < $configData(maxreq))) ||
         (!$configData(maxreq))) &&
        ([onchan $fromnick $chan]) && (![onchansplit $fromnick $chan]) &&
        ([isop $fromnick $chan]) && ([islinked $bot]) &&
        ((($configData(checkaskhost)) &&
          ([matchattr [nick2hand $fromnick $chan] o|o $chan])) ||
         ([matchattr $bot o|o $chan]))} then {
      putbot $bot "yesiwantops $chan $botnick [string trimleft [lindex [split $botname !] 1] "~+-^="]"
      if {$configData(maxreq)} then {
        incr OpRequests($chanLower)
        killUtimerCommand [namespace code "ResetOpRequests $chanLower"]
        utimer 30 [namespace code [list ResetOpRequests $chanLower]]
      }
      if {$configData(log) >= 2} then {
        wpLog o * [mc "requested ops from %s on %s" $bot $chan]
      }
    }
  }
  return 0
}

##
## Set OpRequests for a given channel to 0
##
## Args:    channel
## Returns: nothing
##
proc ResetOpRequests {chan} {
  variable OpRequests

  set OpRequests($chan) 0
  return
}

##
## Answer a request to a request if this bot wants ops
##
## Args:    bot name, command, arg
## Returns: 0
##
proc bot:yesiwantops {bot cmd arg} {
  global strict-host numversion
  variable configData

  set arg [split $arg]
  set chan [lindex $arg 0]
  set fromnick [lindex $arg 1]
  set fromhost [lindex $arg 2]
  # Make sure this bot is on the channel and has ops
  # Make sure the bot requesting ops is on the channel and isn't netsplit
  # Make sure the bot requesting ops really is a bot and has proper flags
  # Make sure the bot requesting ops is really who it says it is
# FIXME: '(![string compare "" $fromhost]) ||' is the logic slennox's
#         botnetop uses -- is this right or is it unsafe?
  if {([validchan $chan]) && ([botonchan $chan]) && ([botisop $chan]) &&
      ([onchan $fromnick $chan]) && (![onchansplit $fromnick $chan]) &&
      ([matchattr $bot b]) &&
      ([matchattr $bot o|o $chan]) && (![matchattr $bot d|d $chan]) &&
      ((![string compare "" $fromhost]) ||
       (![string compare $fromhost \
          [string trimleft [getchanhost $fromnick $chan] "~+-^="]]))} then {
    if {(![matchattr [nick2hand $fromnick $chan] o|o $chan]) &&
        ([string compare "" $fromhost]) && ($configData(addhost))} then {
      if {!${strict-host}} then {
        set host *![string trimleft [getchanhost $fromnick $chan] "~+-^="]
      } else {
        set host *![getchanhost $fromnick $chan]
      }
      setuser $bot HOSTS $host
      wpLog o * [mc "added host %s to %s" $host $bot]
    }
    set log 0
    if {![isop $fromnick $chan]} then {
      pushmode $chan +o $fromnick
      set log 1
    } elseif {$configData(osync)} then {
      # Eggdrop 1.4+ pushmode won't op someone that already appears to have ops
      if {$numversion < 1040000} then {
        pushmode $chan +o $fromnick
      } else {
        putquick "MODE $chan +o $fromnick"
      }
      set log 1
    }
    if {($log) && ($configData(log) >= 2)} then {
      if {[string compare $fromnick $bot]} then {
        wpLog o * [mc "gave ops to %s (using nick %s) on %s" $bot $fromnick $chan]
      } else {
        wpLog o * [mc "gave ops to %s on %s" $bot $chan]
      }
    }
  }
  return 0
}

##
## Botnet wantkey/wantinvite/wantlimit/wantunban
##
## Args:    bot name, command, arg
## Returns: 0
##
proc bot:wantsin {bot cmd arg} {
  variable configData
  variable WhoNicks
  variable WatchModes

  set arg [split $arg]
  set chan [lindex $arg 0]
  set fromnick [lindex $arg 1]
  if {([validchan $chan]) && ([botisop $chan]) &&
      ([matchattr $bot b]) &&
      ([matchattr $bot fo|fo $chan])} then {
    set fromhost [lindex $arg 2]
    switch -exact -- $cmd {
      wantkey {
        set chanmode [split [getchanmode $chan]]
        if {[regexp -- {[k]} [lindex $chanmode 0]]} then {
          set key [lindex $chanmode 1]
          # Some braindead ircds hide the channel key as '*' in the mode change.
          # This will cause eggdrop to refresh its cached channel mode data.
          # The reply from the server will also trigger raw:324.
          if {(![string compare "*" $key]) && ($configData(fixkey))} then {
            set name "[string tolower $chan],key,$fromnick"
            if {![info exists WatchModes($name)]} then {
# FIXME: expire this after awhile
              set WatchModes($name) $bot
# FIXME: could this cause the bot to flood itself?
              putserv "MODE $chan"
            }
          } else {
            putbot $bot "thekey $chan $key"
            if {$configData(log) >= 1} then {
              wpLog o * [mc "gave key for %s to %s" $chan $bot]
            }
          }
        }
      }
      wantinvite {
# FIXME: Older versions of botnetop will not provide $fromhost
#        Maybe we shouldn't invite bots that don't provide $fromhost
#        when 'icheck' is enabled?
        if {($configData(icheck)) && ([string compare "" $fromhost])} then {
          if {![info exists WhoNicks($fromnick)]} then {
            set WhoNicks($fromnick) "$chan $bot $fromhost"
            putserv "WHO $fromnick"
            utimer 60 [namespace code [list WhoUnset $fromnick]]
          }
        } else {
          InviteNick $chan $fromnick $bot
        }
      }
      wantlimit {
# FIXME: expire this after awhile
        set WatchModes([string tolower $chan],limit,$fromnick) $bot
        pushmode $chan +l [expr [llength [chanlist $chan]] + 1]
        if {$configData(log) >= 1} then {
          wpLog o * [mc "raised limit on %s as requested by %s" $chan $bot]
        }
      }
      wantunban {
        set banlist ""
        foreach chanban [chanbans $chan] {
          foreach {ban from time} $chanban {break}
          if {[string match [string tolower $ban] \
               [string tolower $fromhost]]} then {
# FIXME: expire this after awhile
            set WatchModes([string tolower $chan],unban,$fromnick) $bot
            pushmode $chan -b $ban
            lappend banlist $ban
          }
        }
        if {($configData(log) >= 1) && ([string compare "" $banlist])} then {
          regsub -all -- { } [join $banlist] {, } banlist
          wpLog o * [mc "unbanned %s on %s: %s" $bot $chan $banlist]
        }
      }
    }
  }
  return 0
}

##
## Botnet tryjoin
##
## Args:    bot name, command, arg
## Returns: 0
##
proc bot:tryjoin {bot cmd arg} {
  variable Need
  variable Join
  variable ChanKey

  set arg [split $arg]
  set chan [lindex $arg 0]
  if {([validchan $chan]) && (![botonchan $chan])} then {
    set chanLower [string tolower $chan]

    # We don't need to do this for invites because eggdrop will attempt to
    # join the channel when it recieves the invite notice from the server.
    switch -exact -- $cmd {
      thekey {
        set join key
        set key [lindex $arg 1]
        if {[string compare "" $key]} then {
# FIXME: expire ChanKey($chanLower) after awhile
          set ChanKey($chanLower) $key
          set what "JOIN $chan $ChanKey($chanLower)"
        } else {
          set what "JOIN $chan"
        }
      }
      donelimit {
        set join limit
        if {[info exists ChanKey($chanLower)]} then {
          set what "JOIN $chan $ChanKey($chanLower)"
        } else {
          set what "JOIN $chan"
        }
      }
      doneunban {
        set join unban
        if {[info exists ChanKey($chanLower)]} then {
          set what "JOIN $chan $ChanKey($chanLower)"
        } else {
          set what "JOIN $chan"
        }
      }
    }

    if {([checkActionData Need check $join $chan]) &&
        (![checkActionData Join check $join $chan])} then {
      checkActionData Join set $join $chan
      putserv $what
      utimer 10 [namespace code [list checkActionData Join clear $join $chan]]
    }

# FIXME: add logging for this proc?

  }
  return 0
}

##
## raw 352 (who reply)
##
## Args:    from, keyword, arg
## Returns: 0
##
proc raw:352 {from key arg} {
  variable configData
  variable WhoNicks

  set arg [split $arg]
  set fromnick [lindex $arg 5]
  if {[info exists WhoNicks($fromnick)]} then {
    foreach {chan bot fromhost} $WhoNicks($fromnick) {break}
    unset WhoNicks($fromnick)
    if {![string compare $fromhost \
          [string trimleft [lindex $arg 2]@[lindex $arg 3] "~+-^="]]} then {
      InviteNick $fromnick $chan $bot
    }
  }
  return 0
}

##
## Invite a bot to a channel
##
## Args:    channel, nick, bot
## Returns: nothing
##
proc InviteNick {chan nick bot} {
  variable configData

  putserv "INVITE $nick $chan"

  if {$configData(log) >= 1} then {
    if {[string compare $nick $bot]} then {
      wpLog o * [mc "invited %s (using nick %s) to %s" $bot $nick $chan]
    } else {
      wpLog o * [mc "invited %s to %s" $bot $chan]
    }
  }
  return
}

##
## raw 324 (channel mode)
## Some braindead ircds hide the channel key as '*' in the mode change.
##
## Args:    from, keyword, arg
## Returns: 0
##
proc raw:324 {from key arg} {
  variable configData
  variable WatchModes

  set arg [split $arg]
  set chan [lindex $arg 1]
  set mode [lindex $arg 2]

  set chanLower [string tolower $chan]
  set names [array names WatchModes "$chanLower,key,*"]
  if {[string compare "" $names]} then {

    # Channel has a key
    if {[regexp -- {[k]} $mode]} then {

      # Channel has a limit
      if {[regexp -- {[l]} $mode]} then {
        set what "thekey $chan [lindex $arg 4]"
      } else {
        set what "thekey $chan [lindex $arg 3]"
      }
    } else {
      set what "thekey $chan"
    }

    foreach name $names {
      putbot $WatchModes($name) $what
      if {$configData(log) >= 1} then {
        wpLog o * [mc "gave key for %s to %s" $chan $WatchModes($name)]
      }
    }
    array unset WatchModes "$chanLower,key,*"
  }
  return 0
}

##
## If the given nick exists in WhoNicks, unset it
##
## Args:    nick
## Returns: nothing
##
proc WhoUnset {nick} {
  variable WhoNicks

  if {[info exists WhoNicks($nick)]} then {
    unset WhoNicks($nick)
  }
  return
}

##
## Unset stuff when the bot is removed from a channel
##
## Args:    nick, uhost, hand, chan, msg
## Returns: 0
##
proc part:unsetneed {nick uhost hand chan {msg ""}} {
  if {([isbotnick $nick]) && (![validchan $chan])} then {
    ClearNeed $chan
  }
  return 0
}

##
## Check for new channels every 5 minutes
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:5Minutes {minute hour day month year} {
  variable Need

  foreach chan [channels] {
    CheckNeed $chan
  }
  return 0
}

##
## Check if bot needs to set the need-* options for the given channel
##
## Args:    channel
## Returns: nothing
proc CheckNeed {chan} {
  global numversion
  variable Need
  variable OpRequests
  variable KeyJoin

  set chanLower [string tolower $chan]
  if {![info exists Need($chanLower,set)]} then {

    set OpRequests($chanLower) 0
    set KeyJoin($chanLower) 0

# FIXME: support 1.6.0+ 'bind need'
#    if {$numversion < 1060000} then {
      channel set $chan need-op [namespace code "botneeds $chan op"]
      channel set $chan need-key [namespace code "botneeds $chan key"]
      channel set $chan need-invite [namespace code "botneeds $chan invite"]
      channel set $chan need-limit [namespace code "botneeds $chan limit"]
      channel set $chan need-unban [namespace code "botneeds $chan unban"]
#    }
    set Need($chanLower,set) 1
  }
  return
}

##
## Clear need-* options for the given channel
##
## Args:    channel
## Returns: nothing
##
proc ClearNeed {chan} {
  variable Need
  variable OpRequests
  variable KeyJoin

  set chanLower [string tolower $chan]

  array unset Need "$chanLower,*"
  catch {unset OpRequests($chanLower)}
  catch {unset KeyJoin($chanLower)}

  foreach need "need-op need-invite need-key need-limit need-unban" {
    catch {channel set $chan $need ""}
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  foreach chan [channels] {
    ClearNeed $chan
    CheckNeed $chan
  }
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  delay {integer 20 {} seconds {Maximum number of seconds to delay before asking a non-opped bot if it needs ops. This is used by the bot giving ops. Higher settings help reduce +o flooding when a bot joins. Note that the delay is automatically skipped if there are fewer than 'lowbots' bots opped on the channel. Valid settings: 0 to disable delayed op, otherwise 1 or higher.}}
  lowbots {integer 3 {} bots {Minimum number of bots that must be present in a channel for 'delay' to be activated. Valid settings: 0 to disable checking, otherwise 1 or higher.}}
  maxreq {integer 3 {} bots {Maximum number of bots to request ops from at a time. This is used by the bot requesting ops. This feature can completely eliminate +o flooding when a bot joins, but the bot will request ops from fewer bots. Valid settings: 0 to disable, otherwise 1 or higher.}}
  modeop {switch 0 {0/1} {Make the bot send an op request when it sees a bot become opped on a channel? This will make your bot get ops sooner when another bot is given ops. Valid settings: 1 to enable, 0 to disable.}}
  linkop {switch 1 {0/1} {Make the bot send an op request when a bot links? This will make your bot get ops sooner if linking is delayed, and provides a better chance of the bots opping one another if there are linking problems. Valid settings: 1 to enable, 0 to disable.}}
  icheck {switch 0 {0/1} {Check a bot's user@host before inviting it to a channel? Normally, the bot can sometimes invite the wrong user to +i channel if there are splits occurring. Enabling this check can prevent that. Valid settings: 1 to enable, 0 to disable.}}
  osync {switch 0 {0/1} {Give ops to a bot even if it looks like it's already opped on the channel? Normally, the script will ignore a request for ops from a bot if that bot is already opped. But in some desync situations, a bot may appear to be opped even when it isn't, so you might want to enable this if you often have desync problems. Note that enabling this can result in +o flooding. Valid settings: 1 to enable, 0 to disable.}}
  checkbans {switch 1 {0/1} {Check for and remove any bans matching the bot's nick!user@host when it get ops? Valid settings: 1 to enable, 0 to disable.}}
  fixkey {switch 1 {0/1} {Some braindead ircds hide the channel key as '*' in the mode change. This will cause eggdrop to refresh its cached channel mode data when it sees a mode change of '+k *'. Valid settings: 1 to enable, 0 to disable.}}
  tryinvite {switch 1 {0/1} {Some ircds allow /invite to bypass channel bans, limits, and keys. This will cause eggdrop to try inviting a bot to a channel before making mode changes so it can join faster. Valid settings: 1 to enable, 0 to disable.}}
  checkaskhost {switch 0 {0/1} {Check if the user@host for the bot asking if the bot wants ops on a channel is in this bot's userlist. Don't respond to op solicitations if it isn't found. Valid settings: 1 to enable, 0 to disable.}}
  addhost {switch 0 {0/1} {Let the script automatically add new hostmasks for bots? (e.g. if you change a bot's vhost). This feature doesn't work all the time due to complications with protect-telnet, strict-host, bot host sharing flaws, security concerns, etc., but it's convenient when it works. Valid settings: 1 to enable, 0 to disable.}}
  log {integer {range {0 2}} 2 {} {0/1/2} {By default, botnetop.tcl logs all ops, op requests, invite/key limit/unban requests/responses, and host additions. This setting lets you reduce the amount of logging. Valid settings: 2 to enable all logging, 1 to disable logging of ops/op requests, 0 to disable all logging except host additions.}}
}

# Bind defaults
array set bindDefaults {
  mode:watchmodes {{mode - * {} {}}}
  link:checkneeds {{link - * {} {}}}
  join:askops {{join - * {} {}}}
  bot:reqops {{bot - reqops {} {}}}
  bot:doyawantops {{bot - doyawantops {} {}}}
  bot:yesiwantops {{bot - yesiwantops {} {}}}
  bot:wantsin {{bot - wantkey {} {}} {bot - wantinvite {} {}} {bot - wantlimit {} {}} {bot - wantunban {} {}}}
  bot:tryjoin {{bot - thekey {} {}} {bot - donelimit {} {}} {bot - doneunban {} {}}}
  raw:352 {{raw - 352 {} {}}}
  raw:324 {{raw - 324 {} {}}}
  part:unsetneed {{part - * {} {}}}
  time:5Minutes {{time - {?0 * * * *} {} {}} {time - {?5 * * * *} {} {}}}
}
