###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## Inspired by CheckPass.tcl for eggdrop1.0m originally written
## by Drago <Drago@Drago.com> [Dec.21.96]
##
## $Id: checkpass.tcl,v 1.51 2005/01/16 11:29:48 tothwolf Exp $
##

# name: checkpass
# version: 1.20
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: checkpass.conf
# description: Password checking utility that adds a dcc command called \
#	`.checkpass', performs password checking on join for defined \
#	channels, and allows the deletion of users after a set number of \
#	days if they have not set a password.
# requires: wpLog commandChar putnot daysLeft grammarPlural randstring \
#	validHandle findChanFlag findHandOnChannels convertSeconds \
#	moveFields checkChannelSetting listChannelsWithSetting \
#	listUsersWithoutPass listUsersAndHosts formatTime


##
## Check for users with empty passwords: dcc
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:checkpass {hand idx arg} {
  set arg [split [string trimright $arg]]
  set mask [lindex $arg 0]
  if {[string compare "" $mask]} then {
    if {[regexp -- {[+-|&]} $mask]} then {
      # Flags specified
      set type flags
      set desc " with flags '$mask'"
      set userlist [userlist $mask]
    } else {
      # Pattern specified
      set type pattern
      set desc " matching pattern '$mask'"
      set userlist [listUsersAndHosts $mask]
    }
  } else {
    # No mask specified
    set desc ""
    set userlist [userlist]
  }
  # Default won't trigger this
  if {([info exists type]) && \
      (![string compare "" $userlist])} then {
    putdcc $idx "No users or bots matched $type: $mask"
  } else {
    set count 0
    foreach user $userlist {
      if {[passwdok $user ""]} then {
        if {![matchattr $user b]} then {
          append list "$user, "
        } else {
          append list "$user (bot), "
        }
        incr count
      }
    }
    if {$count} then {
      putdcc $idx "Users${desc} without a password ($count): [string trimright $list ", "]"
    } else {
      putdcc $idx "All users${desc} have a password set."
    }
  }
  return 1
}

##
## Check for users with empty passwords when they join channels
##
## Args:    nick, user@host, handle, channel
## Returns: 0
##
proc join:checkpass {nick uhost hand chan} {
  variable configData
  variable CheckUsers

  if {(![isbotnick $nick]) && ([validHandle $hand]) && \
      ([passwdok $hand ""]) && \
      ([checkChannelSetting $chan "+$configData(udefflag)"])} then {
    if {($configData(dkrandpass)) && \
        (([matchattr $hand dk|dk $chan]) || \
         ([findChanFlag $hand dk [listChannelsWithSetting "+$configData(udefflag)"]]))} then {
      setuser $hand PASS [randstring 8]
      setuser $hand COMMENT "checkpass: Set random password: [formatTime [clock seconds]] (d/k flags)"
      wpLog o * "Set random password for 'd' or 'k' flagged user: $nick!$uhost ($hand)"
    } elseif {![info exists CheckUsers($hand)]} then {
      set CheckUsers($hand) "[clock seconds] 0"
      wpLog o * "$nick!$uhost ($hand) does not have a password set, sending a notice..."
      if {$configData(joindelay)} then {
        utimer $configData(joindelay) [namespace code [list SendNotice $nick $hand]]
      } else {
        SendNotice $nick $hand
      }
    }
  }
  return 0
}

##
## Annoy people who refuse to set a password
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:minutely {minute hour day month year} {
  variable configData
  variable CheckUsers

  # Woohoo! We get to annoy users! This is fun!
  if {$configData(annoy)} then {
    set foundcount 0
    if {([string compare "" [set handles [listUsersWithoutPass -b]]]) && \
        ([string compare "" [set channels [listChannelsWithSetting "+$configData(udefflag)"]]])} then {
      foreach hand $handles {
        if {($configData(dkrandpass)) && \
            (([matchattr $hand dk]) || \
             ([findChanFlag $hand dk $channels]))} then {
          # User has d or k flags and 'dkrandpass' is enabled
          setuser $hand PASS [randstring 8]
          setuser $hand COMMENT "checkpass: Set random password: [formatTime [clock seconds]] (d/k flags)"
          wpLog o * "Set random password for 'd' or 'k' flagged user: $hand"
          if {[info exists CheckUsers($hand)]} then {
            unset CheckUsers($hand)
          }
          continue
        }
        foreach chan $channels {
          # Whoop! Found a user without a password whos online...
          if {[handonchan $hand $chan]} then {
            set nick [hand2nick $hand $chan]
            if {$configData(checkidle)} then {
              set idletime [getchanidle $nick $chan]
              # Only increase foundcount once per handle
              if {[info exists foundusers($hand)]} then {
                set found 0
                set index 0
                # Search foundusers to see if 'nick' is already in the list
                # If so, and the idletime is smaller, replace that index
                foreach tmpdata $foundusers($hand) {
                  foreach {tmpnick tmpidle} $tmpdata {break}
                  if {(![string compare $nick $tmpnick]) && \
                      ($idletime < $tmpidle)} then {
                    set foundusers($hand) [lreplace $foundusers($hand) $index $index [list $nick $idletime]]
                    set found 1
                    break
                  }
                  incr index
                }
                if {!$found} then {
                  lappend foundusers($hand) [list $nick $idletime]
                }
              } else {
                lappend foundusers($hand) [list $nick $idletime]
                incr foundcount
              }
              # Don't break here, we want to check idle time on each channel
            } else {
              lappend foundusers($hand) [list $nick 0]
              incr foundcount
              break
            }
          }
        }
      }
    }
    if {$foundcount} then {
      set delay [expr $foundcount * [convertSeconds from minutes $configData(delaytime)]]
      if {$delay < $configData(mindelay)} then {
        set delay $configData(mindelay)
      }
      set time [clock seconds]
      foreach {hand data} [array get foundusers] {
        # This must be clean for each handle
        if {[info exists chanidle]} then {
          unset chanidle
        }
        # Find the least idle nick associated with a handle
        foreach tmpdata $data {
          foreach {tmpnick tmpidle} $tmpdata {break}
          if {(![info exists chanidle]) || \
              ($tmpidle < $chanidle)} then {
            set chanidle $tmpidle
            set nick $tmpnick
          }
        }
        if {[info exists CheckUsers($hand)]} then {
          foreach {seconds idletime} $CheckUsers($hand) {break}
          # If settings permit, don't send notices to idle users
          if {((!$configData(checkidle)) || ($chanidle <= $idletime)) && \
              ([expr $time - $seconds] >= $delay)} then {
            SendNotice $nick $hand
            set CheckUsers($hand) "$time $chanidle"
            break
          } elseif {$chanidle > $idletime} then {
            # Update idletime only
            set CheckUsers($hand) "$seconds $chanidle"
          }
        } else {
          set CheckUsers($hand) "$time $chanidle"
          SendNotice $nick $hand
          break
        }
      }
      if {[string compare "" [set users [array names CheckUsers]]]} then {
        # Clear invalid handles from CheckUsers
        foreach hand $users {
          if {![info exists foundusers($hand)]} then {
            # Didn't find user on any channels
            unset CheckUsers($hand)
          }
        }
      }
    } elseif {[info exists CheckUsers]} then {
      # Clear the entire array, no users with empty passwords were found
      unset CheckUsers
    }
  } else {
    # Check and clean the CheckUsers array even if the main check is disabled
    if {[string compare "" [set userdata [array get CheckUsers]]]} then {
      set time [clock seconds]
      foreach {user data} $userdata {
        if {[expr $time - [lindex $data 0]] >= $configData(mindelay)} then {
          unset CheckUsers($user)
        }
      }
    }
  }
  return 0
}

##
## Find bots without passwords
## Delete users who still don't have a password set
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:daily {minute hour day month year} {
  global owner botnet-nick
  variable configData

  if {$configData(noteowner) == 1} then {
    set notify [userlist n]
  } elseif {$configData(noteowner) == 2} then {
    regsub -all -- {,} $owner {} notify
  }
  set time [clock seconds]
  foreach hand [userlist -b] {
    set passtime [getuser $hand XTRA $configData(timefield)]
    if {($configData(passdays)) && ([passwdok $hand ""])} then {
      if {[string compare "" $passtime]} then {
        if {[expr $time - $passtime] >= [convertSeconds from days $configData(passdays)]} then {
          deluser $hand
          append deluserslist "$hand, "
        }
      } else {
        setuser $hand XTRA [string toupper $configData(timefield)] $time
      }
    } elseif {[string compare "" $passtime]} then {
      setuser $hand XTRA $configData(timefield)
    }
  }
  if {[info exists deluserslist]} then {
    set log "Deleted users: [string trimright $deluserslist ,]. No password was set after [grammarPlural $configData(passdays) day]."
    wpLog o * $log
    if {[info exists notify]} then {
      foreach to $notify {
        sendnote ${botnet-nick} $to $log
      }
    }
  }
  if {[string compare "" [set handles [listUsersWithoutPass b]]]} then {
    regsub -all -- { } $handles {, } list
    if {$configData(botrandpass)} then {
      foreach hand $handles {
        setuser $hand PASS [randstring 8]
        setuser $hand COMMENT "checkpass: Set random password: [formatTime [clock seconds]] (bot)"
      }
      wpLog o * "Random passwords have been set for the following bots: $list"
      if {[info exists notify]} then {
        foreach to $notify {
          sendnote ${botnet-nick} $to "Random passwords have been set for the following bots: $list"
        }
      }
    } else {
      wpLog o * "The following bots have no password set: $list"
      if {[info exists notify]} then {
        foreach to $notify {
          sendnote ${botnet-nick} $to "The following bots have no password set: $list"
        }
      }
    }
  }
  return 0
}

##
## Send a notice about setting a password to the given nick
##
## Args:    nickname, handle
## Returns: nothing
##
proc SendNotice {nick hand} {
  variable configData

  set passtime [getuser $hand XTRA $configData(timefield)]
  if {$configData(passdays)} then {
    if {![string compare "" $passtime]} then {
      set passtime [clock seconds]
      setuser $hand XTRA [string toupper $configData(timefield)] $passtime
    }
  } elseif {[string compare "" $passtime]} then {
    set passtime ""
    setuser $hand XTRA $configData(timefield)
  }
  putnot $nick "You do not have a password set."
  if {[string compare "" $passtime]} then {
    putnot $nick "You access will be removed in [grammarPlural \
      [daysLeft $passtime $configData(passdays)] day] unless you set one."
  }
  set cmdchar [commandChar msg]
  putnot $nick "Please set one now, by typing: ${cmdchar}pass <your password>"
  putnot $nick "For more details: ${cmdchar}help pass"
  return
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      timefield {
        moveFields $oldConfigData(timefield) [string toupper $configData(timefield)]
      }
      udefflag {
        renudef flag $oldConfigData(udefflag) $configData(udefflag)
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  setudef flag $configData(udefflag)
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  annoy {switch 1 {} {0/1} {enable minutely check of userlist and searching of channels?}}
  botrandpass {switch 1 {} {0/1} {set a random password for bots that have no password?}}
  checkidle {switch 1 {} {0/1} {don't send notices to idle users.}}
  delaytime {integer 1 {} minutes {minutes to wait between sending password notices to different users}}
  dkrandpass {switch 1 {} {0/1} {set a random password for a +d or +k user?}}
  joindelay {integer 5 {} seconds {seconds to wait before sending password notice when someone joins a channel}}
  mindelay {integer 300 {} seconds {minimum number of seconds between sending password notices}}
  noteowner {integer 1 {} {0/1/2} {0 - don't send any notes. 1 - send notes to all owners. 2 - send notes to permanent owners only. Send note to owner(s) when deleting users or if bots with no password are found?}}
  passdays {integer 10 {} days {number of days before a user with an empty password is deleted}}
  timefield {word passtime {} field {field for storing check timestamp}}
  udefflag {word checkpass {} flag {flag for enabling/disabling checkpass on channels}}
}

# Bind defaults
array set bindDefaults {
  dcc:checkpass {{dcc m|m checkpass {} {}}}
  join:checkpass {{join -|- * {} {}}}
  time:minutely {{time -|- {* * * * *} {} {}}}
  time:daily {{time -|- {00 00 * * *} {} {}}}
}

# Help Data
array set helpData {
  dcc:checkpass {{[flags|mask]} {shows users and bots that do not have a password set}}
}
