###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: compat.tcl,v 1.19 2005/01/14 06:30:19 tothwolf Exp $
##

# name: compat
# version: 0.7
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# description: Compatibility functions for older versions of eggdrop that \
#	allow modules the ability to use new commands, but still remain \
#	backwards compatible with older eggdrop releases.
# requires: badArgs rangeList rangerList escapeRegexp idx2type \
#	string2string moduleBindUnbind

##
## erasenotes
##   added: 1.3.15
##
#  erasenotes <user> <numberlist>
#    erases some or all stored notes for a user
#    returns: -1 if no such user, -2 if notefile failure, 0 if
#             no such note, or number of erased notes.
#             'erasenote mynick -' erase all notes for mynick.
##
# FIXED: 13 Jan 03 By PPSlim
# NOTE:  It simply returns how many noted where deleted
# If a note index is given, that doesn't exist, it continues
# without error. This doesn't affect the operation.
##
proc erasenotes {args} {
  global notefile

  badArgs $args 2 2 "erasenotes handle noteslist#"

  set user [lindex $args 0]
  set index [lidnex $args 1]
  if {![validuser $user]} then {
    return -1
  }
  if {(![info exists notefile]) || (![file exists $notefile]) || \
      (![file isfile $notefile]) || (![file writable $notefile])} then {
    return -2
  }
  if {[catch {set fd [open $notefile r]}]} then {
    return -2
  } else {
    while {![eof $fd]} {
      lappend notes [gets $fd]
    }
    close $fd

    # from to unixtime body
    # erase matching notes from $notes
    # remove last to first
    # set notes [lreplace $notes $index $index]
    # '-' removes all notes for that user
    # return # removed, or 0 if no match

    set noteuser [string tolower $user]
    set handnotes ""
    set handnotescount 0
    set keepnotes ""
    set keepnotescount 0
    set othernotes ""

    # Remove all notes for user X first, and store then seperate
    foreach note $notes {
      if {![string compare $noteuser [string tolower [lindex $note 1]]]} then {
        lappend handnotes $note
        incr handnotescount
      } else {
        lappend othernotes $note
      }
    }

    foreach noteindex [rangerList [rangeList $index $handnotescount] $handnotescount] {
      lappend keepnotes [lindex $handnotes $noteindex]
      incr keepnotescount
    }

    set notes [concat $othernotes $keepnotes]

    if {[catch {set fd [open $notefile w]}]} then {
      return -2
    } else {
      foreach note $notes {
        puts $fd $note
      }
      close $fd
    }

    return [expr $handnotescount - $keepnotescount]
  }
}

##
## listnotes
##   added: 1.3.15
##
#  listnotes <user> <numberlist>
#    lists existing notes according to the numberlist (ex: "2-4;8;16-")
#    returns: -1 if no such user, -2 if notefile failure, 0 if no
#             such note, list of existing notes.
##
# FIXED: 13th Jan 03 By PPSlim
# NOTE: Returns an empty list on no notes. Simple change if need be
# A lot of it is what rocks your boat. I need more info, to make it
# official, so wait for my mail.
##
proc listnotes {args} {
  global notefile

  badArgs $args 2 2 "listnotes handle noteslist#"

  set user [lindex $args 0]
  set index [lindex $args 1]
  if {![validuser $user]} then {
    return -1
  }
  if {(![info exists notefile]) || (![file exists $notefile]) || \
      (![file isfile $notefile]) || (![file readable $notefile])} then {
    return -2
  }
  if {[catch {set fd [open $notefile r]}]} then {
    return -2
  } else {
    while {![eof $fd]} {
      lappend notes [gets $fd]
    }
    close $fd

    # from to unixtime body
    # list matching notes in $notes
    # '-' lists all notes
    # 1 2 3 etc

    set noteuser [string tolower $user]
    set handnotes ""
    set handnotescount 0
    set result ""

    foreach note $notes {
      if {![string compare [string tolower [lindex $note 1]] $noteuser]} then {
        lappend handnotes $note
        incr handnotescount
      }
    }

    foreach noteindex [rangeList $index $handnotescount] {
      lappend result [lindex $handnotes $noteindex]
    }

    return $result
  }
}

##
## botonchan
##   added: 1.3.24
##
## Args:    channel
## Returns: 1 if the bot is on the given channel
##          0 otherwise
##
proc botonchan {args} {
  global botnick

  badArgs $args 1 1 "botonchan channel"

  set chan [lindex $args 0]
  if {![validchan $chan]} then {
    error "illegal channel: $chan"
  } elseif {[onchan $botnick $chan]} then {
    return 1
  }
  return 0
}

##
## putquick
##   added: 1.3.24
##
## Args:    text
## Returns: nothing
##
proc putquick {args} {
  badArgs $args 1 1 "putquick text"

  putserv [lindex $args 0]
  return
}

##
## botisvoice
##   added: 1.3.25
##
## Args:    channel
## Returns: 1 if the bot is voice (+v) on the given channel
##          0 otherwise
##
proc botisvoice {args} {
  global botnick

  badArgs $args 1 1 "botisvoice channel"

  set chan [lindex $args 0]
  if {![validchan $chan]} then {
    error "illegal channel: $chan"
  } elseif {[isvoice $botnick $chan]} then {
    return 1
  }
  return 0
}

##
## islinked
##   added: 1.3.25
##
## Args:    bot
## Returns: 1 if the given bot is linked to the botnet
##          0 otherwise
##
proc islinked {args} {
  badArgs $args 1 1 "islinked bot"

  if {[lsearch -exact [string tolower [bots]] \
       [string tolower [lindex $args 0]]] != -1} then {
    return 1
  }
  return 0
}

##
## isbotnick
##   added: 1.3.28
##
## Args:    nick
## Returns: 1 if the given nick matches the bot's current nickname
##          0 otherwise
##
proc isbotnick {args} {
  global botnick

  badArgs $args 1 1 "isbotnick nick"

  if {![string compare [string tolower $botnick] \
        [string tolower [lindex $args 0]]]} then {
    return 1
  }
  return 0
}

##
## duration
##   added: 1.3.25, buggy before 1.3.28
##
## Args:    seconds
## Returns: the number of seconds in years, weeks, days, hours, minutes,
##          and seconds
##
proc duration {args} {
  badArgs $args 1 1 "duration seconds"

  set result ""
  set sec [lindex $args 0]
  if {$sec >= 31536000} then {
    set tmp [expr $sec / 31536000]
    if {$tmp > 1} then {
      append result "$tmp years "
    } else {
      append result "$tmp year "
    }
    set sec [expr $sec - ($tmp * 31536000)]
  }
  if {$sec >= 604800} then {
    set tmp [expr $sec / 604800]
    if {$tmp > 1} then {
      append result "$tmp weeks "
    } else {
      append result "$tmp week "
    }
    set sec [expr $sec - ($tmp * 604800)]
  }
  if {$sec >= 86400} then {
    set tmp [expr $sec / 86400]
    if {$tmp > 1} then {
      append result "$tmp days "
    } else {
      append result "$tmp day "
    }
    set sec [expr $sec - ($tmp * 86400)]
  }
  if {$sec >= 3600} then {
    set tmp [expr $sec / 3600]
    if {$tmp > 1} then {
      append result "$tmp hours "
    } else {
      append result "$tmp hour "
    }
    set sec [expr $sec - ($tmp * 3600)]
  }
  if {$sec >= 60} then {
    set tmp [expr $sec / 60]
    if {$tmp > 1} then {
      append result "$tmp minutes "
    } else {
      append result "$tmp minute "
    }
    set sec [expr $sec - ($tmp * 60)]
  }
  if {$sec} then {
    if {$sec > 1} then {
      append result "$sec seconds"
    } else {
      append result "$sec second"
    }
  }
  return [string trimright $result]
}

##
## putkick
##   added 1.4.0  [Note: this isn't 100% emulation]
##
## Args:    channel nick?s? ?comment?
## Returns: nothing
##
proc putkick {args} {
  badArgs $args 3 4 "putkick channel nick?s? ?comment?"

  set chan [lindex $args 0]
  if {![validchan $chan]} then {
    error "illegal channel: $chan"
  } elseif {![botisop $chan]} then {
    error "need op"
  }
  set comment [lidnex $args 2]
  if {![string compare "" $comment]} then {
    set comment ""
  }
  foreach nick [split [lindex $args 1] ,] {
    if {[onchan $nick $chan]} then {
      putquick "KICK $chan $nick $comment"
    }
  }
  return
}

##
## Create wrapper procs for compat commands
##
## Args:    none
## Returns: nothing
##
proc CreateWrapperProcs {} {
  global numversion
  variable CommandList
  variable compatCommandList

  foreach {compatVersion compatCommands} [array get CommandList] {
    if {$numversion < $compatVersion} then {
      foreach command $compatCommands {
        lappend compatCommandList $command
      }
    }
  }

  if {[info exists compatCommandList]} then {
    set namespaceScript "namespace eval :: \{\n"
# FIXME: optimize 'eval' usage
    foreach command $compatCommandList {
      append namespaceScript "proc $command {args} \{\n  return \[eval [namespace code $command] \$args\]\n\}\n"
    }
    append namespaceScript "\n\}"
    if 1 $namespaceScript
  }
  return
}

##
## Remove wrapper procs for compat commands
##
## Args:    none
## Returns: nothing
##
proc RemoveWrapperProcs {} {
  variable compatCommandList

  if {[info exists compatCommandList]} then {
    set namespaceScript "namespace eval :: \{\n"
    foreach command $compatCommandList {
      append namespaceScript "rename $command {}\n"
    }
    append namespaceScript "\n\}"
    if 1 $namespaceScript
  }
  return
}

##
## Workaround for pre 1.4.3 $lastbind partial matches
##
## Args:    idx, text
## Returns: text
##
proc filt:fixlastbind {idx text} {
  variable Lastbindtype

  set Lastbindtype [string2string [idx2type $idx] "{CHAT dcc} {FILES fil}"]
  
  # Don't do any processing to $text
  return $text
}

##
## Workaround for pre 1.4.3 $lastbind partial matches
##
## Args:    name1, name2, op
## Returns: nothing
##
proc TraceLastbind {name1 name2 op} {
  global lastbind
  variable Lastbindtype
  
  if {([string compare "" $Lastbindtype]) &&
      ([string compare "" $lastbind])} then {

# FIXME: [binds] does not exist prior to 1.4.5?
    set binds [binds]
    if {[string compare "" $binds]} then {
      foreach bind $binds {

        foreach {type flags name hits proc} $bind {break}

        # Only 'dcc' and 'fil' bind types can have partial $lastbind matches
        if {[string compare $Lastbindtype $type]} then {
          continue
        }

        # We should never have an ambiguous match because
        # eggdrop won't set $lastbind if one exists
        if {[regexp -nocase -- "^([escapeRegexp $lastbind])+" $name]} then {

          # This won't re-trigger the variable trace
          set lastbind $name
          set Lastbindtype ""
          
          # Found full match, break out of loop
          break
        }
      }
    }
  }
  return
}

##
## Install/remove workaround for pre 1.4.3 $lastbind partial matches
##
## Args:    none
## Returns: nothing
##
proc InstallLastbindFix {} {
  global numversion lastbind
  variable ModuleName

  if {$numversion < 1040300} then {
    # FIXME: bind/unbind code is ugly, change 'lastbindfix' to an option
    moduleBindUnbind bind $ModuleName "option noauto" lastbindfix
    trace variable lastbind w [namespace code TraceLastbind]
  }
  return
}

##
## Remove workaround for pre 1.4.3 $lastbind partial matches
##
## Args:    none
## Returns: nothing
##
proc RemoveLastbindFix {} {
  global numversion lastbind

  if {$numversion < 1040300} then {
    # FIXME: bind/unbind code is ugly, change 'lastbindfix' to an option
    moduleBindUnbind unbind $ModuleName "option noauto" lastbindfix
    trace vdelete lastbind w [namespace code TraceLastbind]
  }
  return
}

##
## Workaround for '[info exists varname]' bug in pre 1.3.28 bots
##
## Arg:     none
## Returns: nothing
##
proc FixVariables {} {
  global numversion
  variable VariableList

  foreach {var ver} [array get VariableList] {
    if {([string compare "" $ver]) || ($numversion < $ver)} then {
      global $var

      catch {set $var}
    }
  }
  return
}

# version "command list"
array set CommandList {
  1031500 "erasenotes listnotes"
  1032400 "botonchan putquick"
  1032500 "botisvoice islinked"
  1032800 "duration isbotnick"
  1040000 "putkick"
}

# FIXME: add other version info
# variable "version"
array set VariableList {
  botname {}
  botnick {}
  config {}
  handlen 1060800
  isjuped 1050200
  lastbind {}
  numversion {}
  server {}
  server-online {}
  uptime {}
  version {}
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  global argv0 numversion

  # Don't do anything when running with tclsh
  if {![info exists argv0]} then {

    # Workarounds for buggy Tcl variables in pre 1.3.28 bots
    catch {set numversion}
    if {$numversion < 1032800} then {
      FixVariables
    }

    # Create wrapper procs for compat commands
    CreateWrapperProcs

    # Eggdrop versions earlier then 1.4.3 show partial matches in $lastbind
    InstallLastbindFix
  }
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  global argv0 numversion

  # Don't do anything when running with tclsh
  if {![info exists argv0]} then {

    # Remove wrapper procs for compat commands
    RemoveWrapperProcs

    # Eggdrop versions earlier then 1.4.3 show partial matches in $lastbind
    RemoveLastbindFix
  }
}

# Bind defaults
array set bindDefaults {
  filt:fixlastbind {{filt -|- * {noauto lastbindfix} {}}}
}
