###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: dcctools.tcl,v 1.13 2005/01/04 03:36:43 tothwolf Exp $
##

# name: dcctools
# version: 1.0
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# description: DCC connection functions for use by other modules.
# provides: handOnParty findHandOnParty idxOwner idx2uhost idx2type \
#	idxistype idxtype2bindtype countControlProc countControlProcHandle \
#	countControlProcUhost listListenPorts listenPortType


##
## Check if the given handle is somewhere on the partyline
##
## Args:    handle
## Returns: 1 if the given handle is on the partyline
##          0 otherwise
##
proc handOnParty {hand} {
  set hand [string tolower $hand]
  foreach {whohand bot uhost flag idle away chan} [string tolower [whom *]] {
    if {![string compare $hand $whohand]} then {
      return 1
    }
  }
  return 0
}

##
## Find the given handle somewhere on the partyline
##
## Args:    handle
## Returns: channel given handle is in on the partyline
##          nothing otherwise
##
proc findHandOnParty {hand} {
  set hand [string tolower $hand]
  foreach {whohand bot uhost flag idle away chan} [whom *] {
    if {![string compare $hand [string tolower $whohand]]} then {
      return $i
    }
  }
  return
}

##
## Check if the given handle is the owner of the given idx
##
## Args:    handle, idx
## Returns: 1 if the given handle is the owner of the given idx
##          0 otherwise
##
proc idxOwner {cmphand cmpidx} {
  foreach {idx hand uhost type extra time} [dcclist] {
    if {(![string compare $cmpidx $idx]) && \
        (![string compare $cmphand $hand])} then {
      return 1
    }
  }
  return 0
}

##
## Find user@host for the given idx
##
## Args:    idx
## Returns: user@host for the given idx
##
proc idx2uhost {cmpidx} {
  foreach {idx hand uhost type extra time} [dcclist] {
    if {![string compare $cmpidx $idx]} then {
      return $uhost
    }
  }
  return
}

##
## Find dcc 'type' for the given idx
##
## Args:    idx
## Returns: 'type' for the given idx
##
proc idx2type {cmpidx} {
  foreach {idx hand uhost type extra time} [dcclist] {
    if {![string compare $cmpidx $idx]} then {
      return $type
    }
  }
  return
}

##
## Find the given type and see if it matches the given type
##
## Args:    idx, type
## Returns: 
##
proc idxistype {cmpidx cmptype} {
  foreach {idx hand uhost type extra time} [dcclist] {
    if {![string compare $cmpidx $idx]} then {
      if {![string compare [string toupper $cmptype] $type]} then {
        return 1
      }
      break
    }
  }
  return 0
}

##
## Return the bind type for the given idx type
##
## Args:    idx type
## Returns: bind type
##
proc idxtype2bindtype {type} {
  switch -exact -- [string toupper $type] {
    BOT {
      return bot
    }
    CHAT {
      return dcc
    }
    FILES {
      return fil
    }
  }
  return
}

##
## Count number of users in a 'control' proc
##
## Args:   proc name
## Returns: number of users in the given control proc
##
proc countControlProc {proc} {
  set count 0
  foreach {idx hand uhost type extra time} [dcclist SCRIPT] {
    if {![string compare $proc [lindex $extra 1]]} then {
      incr count
    }
  }
  return $count
}

##
## Count number of users in a 'control' proc for a handle
##
## Args:   proc name, handle
## Returns: number of users in the given control proc for a handle
##
proc countControlProcHandle {proc cmphand} {
  set count 0
  foreach {idx hand uhost type extra time} [dcclist SCRIPT] {
    if {(![string compare $proc [lindex $extra 1]]) && \
        (![string compare $cmphand $hand])} then {
      incr count
    }
  }
  return $count
}

##
## Count number of users in a 'control' proc for a user@host
##
## Args:   proc name, user@host
## Returns: number of users in the given control proc for a user@host
##
proc countControlProcUhost {proc cmpuhost} {
  set count 0
  foreach {idx hand uhost type extra time} [dcclist SCRIPT] {
    if {(![string compare $proc [lindex $extra 1]]) && \
        (![string compare $cmpuhost $uhost])} then {
      incr count
    }
  }
  return $count
}

##
## Returns list of listen ports
## {port type mask} {port type mask} ...
##
## Args:    none
## Returns: list of listen ports
##
proc listListenPorts {} {
  set ports ""
  foreach {idx hand uhost type extra time} [dcclist TELNET] {
    if {[string compare "" [set portType [listenPortType $hand]]]} then {
      lappend ports "[lindex $type 1] $portType $uhost"
    }
  }
  return $ports
}

##
## Convert dcclist description to port type
##
## Args:    port description
## Returns: port type
##
proc listenPortType {type} {
  switch -exact -- $type {
    "(telnet)" {
      return all
    }
    "(users)" {
      return users
    }
    "(bots)" {
      return bots
    }
    "(script)" {
      return script
    }
    default {
      return
    }
  }
}
