###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: debug.tcl,v 1.16 2005/01/14 06:30:19 tothwolf Exp $
##

# name: debug
# version: 1.0
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: debug.conf
# description: Simulation of eggdrop's built-in functions for debugging and \
#	testing modules with tclsh.
# requires: arraySetAll
# provides: checkStart checkEnd checkArgs


# This module should not ever be loaded by eggdrop
if {![info exists ::argv0]} then {
  error "This module is not designed to be loaded by eggdrop."
}


##
## Check commands
##
## Args:    command name, arguments, min args, max args, return, command info
## Returns: return value passed from args
##
proc checkArgs {command argList data} {
  variable configData
  variable Stats

  foreach {argMinCount argMaxCount ret statusData} $data {break}
  foreach {status version comment} $statusData {break}

  if {$configData(showwarnings)} then {
    switch -exact -- $status {
      added {
        puts stderr "Debug: Warning: the '$command' command was added in version $version and is not avaliable prior to this version."
      }
      removed {
        puts stderr "Debug: Error: the '$command' command was removed in version $version: ${comment}."
      }
      modified {
        puts stderr "Debug: Notice: the '$command' command was modified in version $version: ${comment}."
      }
      depreciated {
        puts stderr "Debug: Warning: the '$command' command is depreciated: ${comment}."
      }
      broken {
        puts stderr "Debug: Notice: the '$command' command is broken: ${comment}."
      }
    }
  }

  if {[string compare "" $status]} then {
    incr Stats($status)
  }

  set argListLength [llength $argList]
  if {($argListLength < $argMinCount)} then {
    puts stderr "Debug: ERROR: '$command' requires at least [GrammarPlural $argMinCount argument], but was called with $argListLength."
    incr Stats(argstoofew)
  } elseif {($argMaxCount != "*") && ($argListLength > $argMaxCount)} then {
    puts stderr "Debug: ERROR: '$command' accepts no more than [GrammarPlural $argMaxCount argument], but was called with $argListLength."
    incr Stats(argstoomany)
  }
  if {$configData(showargs)} then {
    puts stderr "Debug: Args: $command: $argList"
  }
  incr Stats(total)
  return $ret
}

##
## Setup checking variables and procs
##
## Args:    none
## Returns: nothing
##
proc checkStart {} {
  variable configData
  variable CheckCommands
  variable CheckVariables
  variable Stats
  variable AddedCommands

  # Set all stats values to 0
  arraySetAll Stats 0

  # Set variables normally set by eggdrop
  foreach {globalName data} [array get CheckVariables] {
    upvar #0 $globalName name

    if {![info exists name]} then {
      set name $data
    }
  }

  # Define Tcl commands normally added by eggdrop
  foreach {name data} [array get CheckCommands] {
    if {![string compare "" [info commands ::$name]]} then {
      set CheckProc "proc ::$name \{args\} \{return \[[namespace code "checkArgs $name"] \$args [list $data]\]\}"
      if 1 $CheckProc
      lappend AddedCommands $name

# FIXME: this isn't quite right
#        'putloglev' and 'time' already exist, so we can't test them this way

    } elseif {($configData(showwarnings)) && \
              ([lsearch -exact $AddedCommands $name] == -1)} then {
      puts stderr "Debug: Warning: '$name' already exists!"
    }
  }
  return
}

##
## Print some stats info after checking
##
## Args:    none
## Returns: nothing
##
proc checkEnd {} {
  variable configData
  variable Stats
  variable StatsDesc

  if {$configData(showstats)} then {
    puts stderr ""
    puts stderr "Eggdrop command summary:"
    puts stderr "  Added: $Stats(added)"
    puts stderr "  Removed: $Stats(removed)"
    puts stderr "  Modified: $Stats(modified)"
    puts stderr "  Depreciated: $Stats(depreciated)"
    puts stderr "  Broken: $Stats(broken)"
    puts stderr "  Total: $Stats(total)"
    puts stderr "  Called with too few arguments: $Stats(argstoofew)"
    puts stderr "  Called with too many arguments: $Stats(argstoomany)"
    puts stderr ""
    puts stderr "Total Tcl commands executed: [info cmdcount]"
    puts stderr ""
  }
  return
}

# We need this here, since we don't want to load texttools before debug
proc GrammarPlural {number text} {
  if {$number != 1} then {
    return "$number ${text}s"
  }
  return "$number $text"
}

# Set default values for stats
array set Stats {
  total 0
  added 0
  removed 0
  modified 0
  depreciated 0
  broken 0
  argstoofew 0
  argstoomany 0
}

# Variables normally set by eggdrop
# FIXME: config file variable need to be added somewhere
array set CheckVariables {
  botname "BotNick!bot@localhost"
  botnick "BotNick"
  config "eggdrop.conf"
  handlen 9
  isjuped 0
  lastbind ""
  numversion "1061401"
  server "localhost:6667"
  server-online 0
  uptime 0
  version "1.6.14+fakebot 1061401 fakebot"

  servers ""
}

# Tcl commands normally added by eggdrop
# command {argMinCount argMaxCount ret status}
# FIXME: this hasn't been fully updated since 1.4.0
array set CheckCommands {
  putserv {1 1 {} {{} {} {}}}
  puthelp {1 1 {} {{} {} {}}}
  putquick {1 1 {} {added 1.3.24 {}}}
  putkick {2 3 {} {added 1.4.0 {}}}
  putlog {1 1 {} {depreciated {} {use 'putloglev o' instead}}}
  putcmdlog {1 1 {} {depreciated {} {use 'putloglev c' instead}}}
  putxferlog {1 1 {} {depreciated {} {use 'putloglev x' instead}}}
  putloglev {3 3 {} {{} {} {}}}
  dumpfile {2 2 {} {broken {} {does not work for files outside text/ dir}}}
  queuesize {0 1 {} {added 1.3.28 {}}}
  clearqueue {1 1 0 {added 1.3.28 {}}}
  countusers {0 0 0 {{} {} {}}}
  validuser {1 1 0 {{} {} {}}}
  finduser {1 1 * {{} {} {}}}
  userlist {0 2 {} {{} {} {}}}
  passwdok {2 2 0 {{} {} {}}}
  getuser {2 4 {} {{} {} {}}}
  setuser {2 4 {} {{} {} {}}}
  chnick {2 2 0 {{} {} {}}}
  chattr {1 3 - {{} {} {}}}
  botattr {1 3 - {{} {} {}}}
  matchattr {2 3 0 {{} {} {}}}
  matchchanattr {3 3 0 {depreciated {} {use matchattr instead}}}
  adduser {1 2 1 {modified 1.4.1 {hostmask argument is optional}}}
  addbot {2 2 1 {{} {} {}}}
  deluser {1 1 1 {{} {} {}}}
  delhost {2 2 1 {{} {} {}}}
  addchanrec {2 2 1 {{} {} {}}}
  delchanrec {2 2 1 {{} {} {}}}
  getchaninfo {2 2 {} {{} {} {}}}
  setchaninfo {3 3 {} {{} {} {}}}
  setlaston {2 3 {} {{} {} {}}}
  isban {1 2 0 {{} {} {}}}
  ispermban {1 2 0 {{} {} {}}}
  isbansticky {1 2 0 {added 1.3.24 {}}}
  newban {3 5 {} {{} {} {}}}
  newchanban {4 6 {} {{} {} {}}}
  killban {1 1 1 {{} {} {}}}
  killchanban {2 2 1 {{} {} {}}}
  stick {1 2 1 {added 1.3.24 {}}}
  unstick {1 2 1 {added 1.3.24 {}}}
  matchban {1 2 0 {{} {} {}}}
  banlist {0 1 {} {{} {} {}}}
  isexempt {1 2 0 {added 1.3.28 {}}}
  ispermexempt {1 2 0 {added 1.3.28 {}}}
  isexemptsticky {1 2 0 {added 1.3.28 {}}}
  newexempt {3 5 {} {added 1.3.28 {}}}
  newchanexempt {4 6 {} {added 1.3.28 {}}}
  killexempt {1 1 1 {added 1.3.28 {}}}
  killchanexempt {2 2 1 {added 1.3.28 {}}}
  stickexempt {1 2 1 {added 1.4.1 {}}}
  unstickexempt {1 2 1 {added 1.4.1 {}}}
  matchexempt {1 2 0 {added 1.3.28 {}}}
  exemptlist {0 1 0 {added 1.3.28 {}}}
  isinvite {1 2 0 {added 1.3.28 {}}}
  isperminvite {1 2 0 {added 1.3.28 {}}}
  isinvitesticky {1 2 0 {added 1.3.28 {}}}
  newinvite {3 5 {} {added 1.3.28 {}}}
  newchaninvite {4 6 {} {added 1.3.28 {}}}
  killinvite {1 1 1 {added 1.3.28 {}}}
  killchaninvite {2 2 1 {added 1.3.28 {}}}
  stickinvite {1 2 1 {added 1.4.1 {}}}
  unstickinvite {1 2 1 {added 1.4.1 {}}}
  matchinvite {1 2 0 {added 1.3.28 {}}}
  invitelist {0 1 0 {added 1.3.28 {}}}
  newignore {3 4 {} {{} {} {}}}
  killignore {1 1 {} {{} {} {}}}
  ignorelist {0 0 {} {{} {} {}}}
  isignore {1 1 0 {{} {} {}}}
  save {0 0 {} {{} {} {}}}
  reload {0 0 {} {{} {} {}}}
  backup {0 0 {} {{} {} {}}}
  getting-users {0 0 0 {{} {} {}}}
  channel {2 3 {} {{} {} {}}}
  savechannels {0 0 {} {{} {} {}}}
  loadchannels {0 0 {} {{} {} {}}}
  channels {0 0 {} {{} {} {}}}
  isbotnick {1 1 0 {added 1.3.28 {}}}
  botisop {1 1 0 {{} {} {}}}
  botisvoice {1 1 0 {added 1.3.25 {}}}
  botonchan {1 1 0 {added 1.3.24 {}}}
  wasop {2 2 0 {added 1.4.2 {}}}
  isop {2 2 0 {{} {} {}}}
  isvoice {2 2 0 {{} {} {}}}
  onchan {2 2 0 {{} {} {}}}
  nick2hand {2 1 0 {modified 1.3.27 {channel argument is optional}}}
  handonchan {2 2 0 {{} {} {}}}
  hand2nick {1 2 {} {modified 1.3.27 {channel argument is optional}}}
  ischanban {2 2 0 {{} {} {}}}
  ischanexempt {2 2 0 {added 1.3.24 {}}}
  ischaninvite {2 2 0 {added 1.3.24 {}}}
  chanbans {1 1 {} {{} {} {}}}
  chanexempts {1 1 {} {added 1.3.24 {}}}
  chaninvites {1 1 {} {added 1.3.24 {}}}
  resetbans {1 1 {} {{} {} {}}}
  resetexempts {1 1 {} {added 1.3.24 {}}}
  resetinvites {1 1 {} {added 1.3.24 {}}}
  resetchan {1 1 {} {{} {} {}}}
  getchanhost {1 2 {} {modified 1.3.27 {channel argument is optional}}}
  getchanjoin {2 2 0 {{} {} {}}}
  onchansplit {2 2 0 {{} {} {}}}
  chanlist {1 2 {} {{} {} {}}}
  getchanidle {2 2 0 {{} {} {}}}
  getchanmode {1 1 {} {{} {} {}}}
  jump {1 3 {} {{} {} {}}}
  pushmode {2 * {} {{} {} {}}}
  flushmode {1 1 {} {{} {} {}}}
  topic {1 1 {} {{} {} {}}}
  validchan {1 1 1 {{} {} {}}}
  isdynamic {1 1 1 {{} {} {}}}
  setudef {2 2 {} {added 1.5.2 {}}}
  renudef {2 2 {} {added 1.5.2 {}}}
  deludef {2 2 {} {added 1.5.2 {}}}
  putdcc {2 2 {} {{} {} {}}}
  dccbroadcast {1 1 {} {{} {} {}}}
  dccputchan {2 2 {} {{} {} {}}}
  boot {1 2 {} {{} {} {}}}
  restart {0 0 {} {{} {} {}}}
  rehash {0 0 {} {{} {} {}}}
  dccsimul {2 2 {} {{} {} {}}}
  hand2idx {1 1 -1 {{} {} {}}}
  idx2hand {1 1 {} {{} {} {}}}
  valididx {1 1 0 {{} {} {}}}
  getchan {1 1 0 {{} {} {}}}
  setchan {2 2 {} {{} {} {}}}
  console {1 3 * {{} {} {}}}
  echo {1 2 0 {{} {} {}}}
  putbot {2 2 {} {{} {} {}}}
  putallbots {1 1 {} {{} {} {}}}
  killdcc {1 1 {} {{} {} {}}}
  bots {0 0 {} {{} {} {}}}
  botlist {0 0 {} {added 1.3.25 {}}}
  islinked {1 1 0 {added 1.3.25 {}}}
  dccused {0 0 5 {{} {} {}}}
  dcclist {0 0 {} {{} {} {}}}
  whom {1 1 {} {{} {} {}}}
  getdccidle {1 1 0 {{} {} {}}}
  getdccaway {1 1 {} {{} {} {}}}
  setdccaway {2 2 {} {{} {} {}}}
  connect {2 2 0 {{} {} {}}}
  listen {2 5 {} {{} {} {}}}
  dccdumpfile {2 2 {} {broken {} {does not work for files outside text/ dir}}}
  bind {3 4 {} {{} {} {}}}
  unbind {4 4 {} {{} {} {}}}
  logfile {0 3 $args {{} {} {}}}
  maskhost {1 1 {} {{} {} {}}}
  timer {2 2 timer0 {{} {} {}}}
  utimer {2 2 utimer0 {{} {} {}}}
  timers {0 0 {} {{} {} {}}}
  utimers {0 0 {} {{} {} {}}}
  killtimer {1 1 {} {{} {} {}}}
  killutimer {1 1 {} {{} {} {}}}
  unixtime {0 0 {} {{} {} {}}}
  duration {1 1 {} {modified 1.3.28 {existed in 1.3.25-1.3.27, but was broken until 1.3.28}}}
  strftime {1 2 {} {{} {} {}}}
  ctime {1 1 {} {{} {} {}}}
  myip {0 0 {} {{} {} {}}}
  rand {1 1 0 {{} {} {}}}
  control {2 2 {} {{} {} {}}}
  sendnote {3 3 1 {{} {} {}}}
  link {0 1 1 {{} {} {}}}
  unlink {1 1 1 {{} {} {}}}
  encrypt {2 2 {} {{} {} {}}}
  decrypt {2 2 {} {{} {} {}}}
  encpass {1 1 {} {added 1.3.27 {}}}
  die {0 1 {} {{} {} {}}}
  unames {0 0 {} {{} {} {}}}
  assoc {1 2 {} {{} {} {}}}
  killassoc {1 1 {} {{} {} {}}}
  setpwd {2 2 {} {{} {} {}}}
  getpwd {1 1 / {{} {} {}}}
  getfiles {1 1 {} {{} {} {}}}
  getdirs {1 1 {} {{} {} {}}}
  dccsend {2 2 0 {{} {} {}}}
  filesend {2 3 1 {{} {} {}}}
  setdesc {3 3 {} {{} {} {}}}
  getdesc {2 2 {} {{} {} {}}}
  setowner {3 3 {} {{} {} {}}}
  getowner {2 2 {} {{} {} {}}}
  setlink {3 3 {} {{} {} {}}}
  getlink {2 2 {} {{} {} {}}}
  getfileq {1 1 {} {{} {} {}}}
  mkdir {1 2 0 {{} {} {}}}
  rmdir {1 1 0 {{} {} {}}}
  mv {2 2 1 {{} {} {}}}
  cp {2 2 1 {{} {} {}}}
  getflags {1 1 {} {{} {} {}}}
  setflags {1 2 0 {{} {} {}}}
  notes {1 2 0 {{} {} {}}}
  erasenotes {2 2 1 {added 1.3.15 {}}}
  listnotes {2 2 {} {added 1.3.15 {}}}
  storenote {4 4 0 {{} {} {}}}
  gethosts {1 1 {} {removed 1.3.0 {use 'getuser HOSTS' instead}}}
  addhost {2 2 {} {removed 1.3.0 {use 'setuser HOSTS' instead}}}
  chpass {2 2 {} {removed 1.3.0 {use 'setuser PASS' instead}}}
  getxtra {1 1 {} {removed 1.3.0 {use 'getuser XTRA' instead}}}
  setxtra {3 3 {} {removed 1.3.0 {no longer relevant}}}
  getinfo {1 1 {} {removed 1.3.0 {use 'getuser INFO' instead}}}
  setinfo {2 2 {} {removed 1.3.0 {use 'setuser INFO' instead}}}
  getaddr {1 1 {} {removed 1.3.0 {use 'getuser BOTADDR' instead}}}
  setaddr {2 2 {} {removed 1.3.0 {use 'setuser BOTADDR' instead}}}
  getdccdir {1 1 {} {removed 1.3.0 {use 'getuser DCCDIR' instead}}}
  setdccdir {2 2 {} {removed 1.3.0 {use 'setuser DCCDIR' instead}}}
  getcomment {1 1 {} {removed 1.3.0 {use 'getuser COMMENT' instead}}}
  setcomment {2 2 {} {removed 1.3.0 {use 'setuser COMMENT' instead}}}
  getemail {1 1 {} {removed 1.3.0 {use 'getuser XTRA EMAIL' instead}}}
  setemail {2 2 {} {removed 1.3.0 {use 'setuser XTRA EMAIL' instead}}}
  getlaston {1 1 {} {removed 1.3.0 {use 'getuser LASTON' instead}}}
  getchanlaston {2 2 {} {removed 1.3.0 {use 'getuser LASTON' instead}}}
  time {0 0 {} {removed 1.4.0 {use 'strftime \"%H:%M\"' instead}}}
  date {0 0 {} {removed 1.4.0 {use 'strftime \"%d %b %Y\"' instead}}}
  setdnloads {3 3 {} {removed 1.4.0 {use 'setuser FSTAT d' instead}}}
  getdnloads {1 1 {} {removed 1.4.0 {use 'getuser FSTAT d' instead}}}
  setuploads {3 3 {} {removed 1.4.0 {use 'setuser FSTAT u' instead}}}
  getuploads {1 1 {} {removed 1.4.0 {use 'getuser FSTAT u' instead}}}
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  checkStart
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  checkEnd
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  showargs {switch 0 {} {0/1} {Print command arguments to stderr?}}
  showwarnings {switch 1 {} {0/1} {Print command warnings to stderr?}}
  showstats {switch 1 {} {0/1} {Print command statistics to stderr?}}
}
