###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: emailtools.tcl,v 1.10 2005/01/04 03:36:43 tothwolf Exp $
##

# name: emailtools
# version: 0.6
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: emailtools.conf
# description: Provides functions other modules can use to send email.
# provides: sendEmail contactAddress mailAddress


##
## Send an email
##
## Args:    to, from, subject, body
## Returns: nothing
## Errors:  if unable to connect to outgoing server
##
proc sendEmail {to from subject body} {
  variable configData
  variable EmailData

  set EmailData(to) $to
  set EmailData(from) $from
  set EmailData(subject) $subject
  set EmailData(body) $body
  if {[catch {set idx [connect $configData(serverhost) $configData(serverport)]} result]} then {
    unset EmailData
    error $result
  } else {
    control $idx [namespace code smtpControl]
  }
  return
}

##
## Communicate with sendmail
##
## Args:    idx, arg
## Returns: 1 if message accepted
##          0 otherwise
##
proc smtpControl {idx arg} {
  variable EmailData

  set arg [split $arg]
  switch -exact -- [lindex $arg 0] {
    220 {
      putdcc $idx "MAIL FROM: $EmailData(from)"
    }
    250 {
      switch -exact -- [string tolower [join [lrange $arg 2 end]]] {
        "sender ok" {
          putdcc $idx "RCPT TO: $EmailData(to)"
        }
        "recipient ok" {
          putdcc $idx "DATA"
        }
        "message accepted for delivery" {
          putdcc $idx "QUIT"
          unset EmailData
          return 1
        }
      }
    }
    354 {
      putdcc $idx "Subject: $EmailData(subject)"
      putdcc $idx $EmailData(body)
      putdcc $idx "."
    }
  }
  return 0
}

##
## Return a contact email address for the user running this code
##
## Args:    none
## Returns: email address
##
proc contactAddress {} {
  global tcl_platform admin
  variable configData

  if {[string compare "" $configData(contactemail)]} then {
    return $configData(contactemail)
  } elseif {([info exists admin]) && \
            ([string compare "" [set address [mailAddress $admin]]])} then {
    return $address
  } elseif {[info exists tcl_platform(user)]} then {
    return "$tcl_platform(user)@[info hostname]"
  }
  return ""
}

##
## -- mailAddress
##
## Argument:     A string possibly containing an email address in
##               RFC 821 format.
## Return value: The found address or an empty string. The latter
##               signals that no address could be found,,
##
## Takes a string and tries to extract an email address contained
## in it. Copied from the Exmh application, file 'lib/msgShow.tcl'.
## Copyright is held by Brent Welch <bwelch@eng.sun.com>. This is
## *no* full-fledged RFC 821 parser, so it may miss valid addresses.
##
proc mailAddress {text} {
  # First remove superfluous whitespace and surrounding
  # parentheses.

  set text [string trim $text]

  if {[regsub {\(.*\)} $text {} newtext]} {
    set text $newtext
  }

  # Now look out for tokens grouped around @,
  # neither whitespace nor '"' allowed in it.

  if {[regexp {[^     "]*@[^  "]*} $text token]} {
    # Remove angle brackets, if present, they are not part of
    # the address but there used to distinguish it from a real
    # name contained in the string.

    set token [string trim $token <>]
  } else {
    # No remove address was found (one containing @),
    # so search for something in angle brackets.
    # If nothing like this is present give up and
    # return an empty string.

    if [regexp {<.*>} $text token] {
      set token [string trim $token <>]
    } else {
      set token ""
    }
  }

  return $token
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  contactemail {string "" {} address {Email address of user running this script}}
  serverhost {word mail {} host {Outgoing email server host}}
  serverport {integer 25 {} port {Outgoing email server port}}
}
