###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: floodchk.tcl,v 1.76 2005/01/04 03:36:43 tothwolf Exp $
##

# name: floodchk
# version: 1.5
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: floodchk.conf
# description: Provides flood/throttle checking for commands
# requires: wpLog utimerExists
# provides: floodCheck checkGlobalFlood checkUserFlood clearGlobalFlood \
#	clearUserFlood expireGlobalFloodData expireUserFloodData \
#	expireFloodData

# Levels supported:
#   global flood (any user, any command)
#   global throttle (any user, specific command)
#   user flood (specific user, any command)
#   user throttle (specific user, specific command)
#
# Data variable formats:
#   GlobalFloodData {count time {{cmdname cmdcount cmdtime} {cmdname cmdcount cmdtime} ...}}
#   UserFloodData(nick!user@host) {count time ignore {{cmdname cmdcount cmdtime} {cmdname cmdcount cmdtime} ...}}


##
## Check and throttle/ignore commands that are sent too fast
##
## Args:    nick, uhost, command, global, checkglobal, checkuser
## Returns: 1 if command should be processed
##          0 otherwise
##
proc floodCheck {nick uhost {command ""} {global 1} {checkglobal 1} {checkuser 1}} {
  set unixtime [clock seconds]
  if {[string compare "" $command]} then {
    set throttle 1
  } else {
    set throttle 0
  }
  if {(!$global) || ([set return [checkGlobalFlood $unixtime $command $checkglobal $throttle]])} then {
    set return [checkUserFlood $unixtime $nick $uhost $command $checkuser $throttle]
  }
  return $return
}

##
## Check for global flood
##
## Args:    unixtime, command
## Returns: 1 if check passed
##          0 otherwise
##
proc checkGlobalFlood {unixtime command checkflood throttle} {
  variable configData
  variable FloodConfig
  variable GlobalFloodData

  if {[info exists GlobalFloodData]} then {
    foreach {count time commands} $GlobalFloodData {break}
    if {$count} then {
      if {($count >= $FloodConfig(gfcount)) && \
          ([expr $unixtime - $time] <= $FloodConfig(gftime))} then {
        set GlobalFloodData [list 0 $unixtime $commands]
        utimer $configData(gfloodtime) [namespace code clearGlobalFlood]
        wpLog o * "Global Flood detected -- ignoring further commands"
        return 0
      } else {
        if {[expr ($time + $FloodConfig(gftime)) - $unixtime] > 0} then {
          if {($checkflood) || (!$throttle)} then {
            incr count
          }
        } else {
          set count 1
          set time $unixtime
        }
        if {($throttle) && ([llength $commands])} then {
          foreach {cmdindex cmdname cmdcount cmdtime} [FindCommand $command $commands] {break}
          if {$cmdindex == -1} then {
            lappend commands [list $command 1 $unixtime]
          } else {
            if {$cmdcount} then {
              if {($cmdcount >= $FloodConfig(gtcount)) && \
                  ([expr $unixtime - $cmdtime] <= $FloodConfig(gttime))} then {
                set commands [lreplace $commands $cmdindex $cmdindex [list $cmdname 0 $unixtime]]
                set GlobalFloodData [list $count $time $commands]
                utimer $configData(uthrottime) [namespace code [list clearGlobalFlood $cmdname]]
                wpLog o * "Global Flood detected for '$cmdname' -- throttled"
                return 0
              } else {
                if {[expr ($cmdtime + $FloodConfig(gttime)) - $unixtime] > 0} then {
                  incr cmdcount
                } else {
                  set cmdcount 1
                  set cmdtime $unixtime
                }
              }
              set commands [lreplace $commands $cmdindex $cmdindex [list $cmdname $cmdcount $cmdtime]]
            } else {
              return 0
            }
          }
        }
        set GlobalFloodData [list $count $time $commands]
      }
    } else {
      return 0
    }
  } else {
    set GlobalFloodData [list 1 $unixtime [list "$command 1 $unixtime"]]
  }
  return 1
}

##
## Check and throttle/ignore commands from someone that are sent too fast
##
## Args:    unixtime, nick, uhost, command, userflood {0, 1}
## Returns: 1 if command should be processed
##          0 otherwise
##
proc checkUserFlood {unixtime nick uhost command checkflood throttle} {
  global botnet-nick
  variable configData
  variable FloodConfig
  variable UserFloodData

  set uhost [string trimleft $uhost ~]
  set who "$nick!$uhost"
  if {[info exists UserFloodData($who)]} then {
    foreach {count time ignore commands} $UserFloodData($who) {break}
    if {$count} then {
      if {($count >= $FloodConfig(ufcount)) && \
          ([expr $unixtime - $time] <= $FloodConfig(uftime))} then {
        set UserFloodData($who) [list 0 $unixtime 0 $commands]
        utimer $configData(ufloodtime) [namespace code [list clearUserFlood $who]]
        wpLog o * "Flood detected from $who -- ignoring further commands"
        return 0
      } else {
        if {[expr ($time + $FloodConfig(uftime)) - $unixtime] > 0} then {
          if {($checkflood) || (!$throttle)} then {
            incr count
          }
        } else {
          set count 1
          set time $unixtime
        }
        if {($throttle) && ([llength $commands])} then {
          foreach {cmdindex cmdname cmdcount cmdtime} [FindCommand $command $commands] {break}
          if {$cmdindex == -1} then {
            lappend commands [list $command 1 $unixtime]
          } else {
            if {$cmdcount} then {
              if {($cmdcount >= $FloodConfig(utcount)) && \
                  ([expr $unixtime - $cmdtime] <= $FloodConfig(uttime))} then {
                set commands [lreplace $commands $cmdindex $cmdindex [list $cmdname 0 $unixtime]]
                set UserFloodData($who) [list $count $time $ignore $commands]
                utimer $configData(uthrottime) [namespace code [list clearUserFlood $who $cmdname]]
                wpLog o * "Flood detected from $who for '$cmdname' -- throttled"
                return 0
              } else {
                if {[expr ($cmdtime + $FloodConfig(uttime)) - $unixtime] > 0} then {
                  incr cmdcount
                } else {
                  set cmdcount 1
                  set cmdtime $unixtime
                }
              }
              set commands [lreplace $commands $cmdindex $cmdindex [list $cmdname $cmdcount $cmdtime]]
            } else {
              return 0
            }
          }
        }
        set UserFloodData($who) [list $count $time $ignore $commands]
      }
    } else {
      if {($configData(ignorecount)) && ($ignore > -1)} then {
        if {$ignore < $configData(ignorecount)} then {
          incr ignore
          set UserFloodData($who) [list $count $time $ignore $commands]
        } else {
          set maskhost [maskhost $nick!*$uhost]
          if {![isignore $maskhost]} then {
            wpLog o * "Flood detected from $who, limit of $configData(ignorecount) additional commands reached -- adding to ignore list"
            newignore $maskhost ${botnet-nick} "Flood from $nick!$uhost" $configData(ignoretime)
          }
          set ignore -1
          set UserFloodData($who) [list $count $time $ignore $commands]
        }
      }
      return 0
    }
  } else {
    set UserFloodData($who) [list 1 $unixtime 0 [list "$command 1 $unixtime"]]
  }
  return 1
}

##
## Clear global flood check data
##
## Args:    type {flood, throttle}, command
## Returns: 1 if the given command was removed
##          0 otherwise
##
proc clearGlobalFlood {{command ""}} {
  variable GlobalFloodData

  if {[info exists GlobalFloodData]} then {
    if {![string compare "" $command]} then {
      if {[info exists GlobalFloodData]} then {
        unset GlobalFloodData
        return 1
      }
    } else {
      foreach {count time commands} $GlobalFloodData {break}
      set index [lindex [FindCommand $command $commands] 0]
      if {$index != -1} then {
        set commands [lreplace $commands $index $index]
        set GlobalFloodData [list $count $time $commands]
        return 1
      }
    }
  }
  return 0
}

##
## Clear user flood check data
##
## Args:    who, command
## Returns: 1 if the given nick!user@host was removed
##          0 otherwise
##
proc clearUserFlood {who {command ""}} {
  variable UserFloodData

  if {[info exists UserFloodData($who)]} then {
    foreach {count time ignore commands} $UserFloodData($who) {break}
    if {![string compare "" $command]} then {
      unset UserFloodData($who)
      return 1
    } else {
      set index [lindex [FindCommand $command $commands] 0]
      if {$index != -1} then {
        set commands [lreplace $commands $index $index]
        set UserFloodData($who) [list $count $time $ignore $commands]
        return 1
      }
    }
  }
  return 0
}

##
## Clear out stale data from global flood checking variable
##
## Args:    none
## Returns: nothing
##
proc expireGlobalFloodData {} {
  variable FloodConfig
  variable GlobalFloodData

  set unixtime [clock seconds]
  if {[info exists GlobalFloodData]} then {
    foreach {count time commands} $GlobalFloodData {break}
    if {($count) && ([expr $unixtime - $time] >= $FloodConfig(gftime)) && \
        (![utimerExists [namespace code clearGlobalFlood]])} then {
      unset GlobalFloodData
    }
  }
  return
}

##
## Clear out stale data from user flood checking variables
##
## Args:    none
## Returns: nothing
##
proc expireUserFloodData {} {
  variable FloodConfig
  variable UserFloodData

  set unixtime [clock seconds]
  foreach who [array names UserFloodData] {
    foreach {count time ignore commands} $UserFloodData($who) {break}
    if {($count) && ([expr $unixtime - $time] >= $FloodConfig(uftime)) && \
        (![utimerExists [namespace code "clearUserFlood $who"]])} then {
      unset UserFloodData($who)
    }
  }
  return
}

##
## Clear out stale data from flood checking variables
##
## Args:    none
## Returns: nothing
##
proc expireFloodData {} {
  variable FloodConfig
  variable GlobalFloodData
  variable UserFloodData

  set unixtime [clock seconds]
  if {[info exists GlobalFloodData]} then {
    foreach {count time commands} $GlobalFloodData {break}
    if {($count) && ([expr $unixtime - $time] >= $FloodConfig(gftime)) && \
        (![utimerExists [namespace code clearGlobalFlood]])} then {
      unset GlobalFloodData
    }
  }
  foreach who [array names UserFloodData] {
    foreach {count time ignore commands} $UserFloodData($who) {break}
    if {($count) && ([expr $unixtime - $time] >= $FloodConfig(uftime)) && \
        (![utimerExists [namespace code "clearUserFlood $who"]])} then {
      unset UserFloodData($who)
    }
  }
  return
}

##
## Display flood checking settings
##
## Args:    idx
## Returns: nothing
##
proc FloodStatusSettings {idx} {
  variable configData
  variable FloodConfig

  putdcc $idx "Flood check settings:"
  putdcc $idx "Global:"
  putdcc $idx "  Throttle count: $FloodConfig(gtcount)"
  putdcc $idx "  Throttle seconds: $FloodConfig(gttime)"
  putdcc $idx "  Throttle ignore time: $configData(gthrottime)"
  putdcc $idx "  Flood count: $FloodConfig(gfcount)"
  putdcc $idx "  Flood seconds: $FloodConfig(gftime)"
  putdcc $idx "  Flood ignore time: $configData(gfloodtime)"
  putdcc $idx "User:"
  putdcc $idx "  Throttle count: $FloodConfig(utcount)"
  putdcc $idx "  Throttle seconds: $FloodConfig(uttime)"
  putdcc $idx "  Throttle ignore time: $configData(uthrottime)"
  putdcc $idx "  Flood count: $FloodConfig(ufcount)"
  putdcc $idx "  Flood seconds: $FloodConfig(uftime)"
  putdcc $idx "  Flood ignore time: $configData(ufloodtime)"
  putdcc $idx "  Excessive command ignore count: $configData(ignorecount)"
  putdcc $idx "  Excessive command ignore time: $configData(ignoretime)"
  return
}

##
## Display flood checking global status
##
## Args:    idx
## Returns: nothing
##
proc FloodStatusGlobal {idx} {
  variable configData
  variable FloodConfig
  variable GlobalFloodData

  if {[info exists GlobalFloodData]} then {
    set unixtime [clock seconds]
    putdcc $idx "Global flood check status:"
    foreach {count time commands} $GlobalFloodData {break}
    if {$count} then {
      set gftime [expr ($time + $FloodConfig(gftime)) - $unixtime]
      putdcc $idx "  Command count: $count"
      putdcc $idx "  Time issued: [clock format $time]"
      if {$gftime > 1} then {
        putdcc $idx "  Time remaining: [duration $gftime]"
      } else {
        putdcc $idx "  Pending expiration."
      }
    } else {
      set gfrtime [expr ($time + $configData(gfloodtime)) - $unixtime]
      putdcc $idx "  Flood triggered"
      putdcc $idx "  Trigger time: [clock format $time]"
      if {$gfrtime > 1} then {
        putdcc $idx "  Time remaining: [duration $gfrtime]"
      } else {
        putdcc $idx "  Pending expiration."
      }
    }
    # putdcc can't handle ""
    putdcc $idx " "
    if {[llength $commands]} then {
      putdcc $idx "  Detailed command status:"
      foreach cmd $commands {
        foreach {cmdname cmdcount cmdtime} $cmd {break}
        putdcc $idx "  Name: $cmdname"
        if {$cmdcount} then {
          set gttime [expr ($cmdtime + $FloodConfig(gttime)) - $unixtime]
          putdcc $idx "    Command count: $cmdcount"
          putdcc $idx "    Time issued: [clock format $cmdtime]"
          if {$gttime > 1} then {
            putdcc $idx "    Time remaining: [duration $gttime]"
          } else {
            putdcc $idx "    Pending expiration."
          }
        } else {
          set gtrtime [expr ($cmdtime + $configData(gthrottime)) - $unixtime]
          putdcc $idx "    Throttle triggered"
          putdcc $idx "    Trigger time: [clock format $cmdtime]"
          if {$gtrtime > 1} then {
            putdcc $idx "    Time remaining: [duration $gtrtime]"
          } else {
            putdcc $idx "    Pending expiration."
          }
        }
      }
    }
  } else {
    putdcc $idx "No global flood check data."
  }
  return
}

##
## Display flood checking user status
##
## Args:    idx
## Returns: nothing
##
proc FloodStatusUser {idx} {
  variable configData
  variable FloodConfig
  variable UserFloodData

  if {[llength [set data [lsort [array names UserFloodData]]]]} then {
    set unixtime [clock seconds]
    putdcc $idx "User flood check status:"
    foreach who $data {
      set whohand [finduser $who]
      if {[string compare * $whohand]} then {
        putdcc $idx "  User: $who ($whohand)"
      } else {
        putdcc $idx "  User: $who"
      }
      foreach {count time ignore commands} $UserFloodData($who) {break}
      if {$count} then {
        set uftime [expr ($time + $FloodConfig(uftime)) - $unixtime]
        putdcc $idx "  Command count: $count"
        putdcc $idx "  Time issued: [clock format $time]"
        if {$uftime > 1} then {
          putdcc $idx "  Time remaining: [duration $uftime]"
        } else {
          putdcc $idx "  Pending expiration."
        }
      } else {
        set ufrtime [expr ($time + $configData(ufloodtime)) - $unixtime]
        putdcc $idx "  Flood triggered"
        putdcc $idx "  Trigger time: [clock format $time]"
        if {$ufrtime > 1} then {
          putdcc $idx "  Time remaining: [duration $ufrtime]"
        } else {
          putdcc $idx "  Pending expiration."
        }
      }
      if {$ignore < 0} then {
        putdcc $idx "  Ignore triggered"
        set whosplit [split $who !]
        set maskhost [maskhost "[lindex $whosplit 0]!*[lindex $whosplit 1]"]
        if {[isignore $maskhost]} then {
          set masklower [string tolower $maskhost]
          if {[llength [set ignores [ignorelist]]]} then {
            foreach ignore $ignores {
              foreach {hostmask comment expire added creator} $ignore {break}
              if {![string compare $masklower [string tolower $hostmask]]} then {
                putdcc $idx "  Ignore mask: $hostmask"
                putdcc $idx "  Ignore comment: $comment"
                putdcc $idx "  Ignore added: [clock format $added]"
                putdcc $idx "  Time since ignore: [duration [expr $unixtime - $added]]"
                # bugfix: buggy expire field
                if {$expire == $added} then {
                  set expire [expr $added + $configData(ignoretime)]
                }
                putdcc $idx "  Ignore expires: [clock format $expire]"
                putdcc $idx "  Time left: [duration [expr $expire - $unixtime]]"
                break
              }
            }
          }
        } else {
          putdcc $idx "  Mask '$maskhost' not found in ignore list."
        }
      }
      # putdcc can't handle ""
      putdcc $idx " "
      if {[llength $commands]} then {
        putdcc $idx "  Detailed command status:"
        foreach cmd $commands {
          foreach {cmdname cmdcount cmdtime} $cmd {break}
          putdcc $idx "  Name: $cmdname"
          if {$cmdcount} then {
            set uttime [expr ($cmdtime + $FloodConfig(uttime)) - $unixtime]
            putdcc $idx "    Command count: $cmdcount"
            putdcc $idx "    Time issued: [clock format $cmdtime]"
            if {$uttime > 1} then {
              putdcc $idx "    Time remaining: [duration $uttime]"
            } else {
              putdcc $idx "    Pending expiration."
            }
          } else {
            set utrtime [expr ($cmdtime + $configData(uthrottime)) - $unixtime]
            putdcc $idx "    Throttle triggered"
            putdcc $idx "    Trigger time: [clock format $cmdtime]"
            if {$utrtime > 1} then {
              putdcc $idx "    Time remaining: [duration $utrtime]"
            } else {
              putdcc $idx "    Pending expiration."
            }
          }
        }
      }
    }
  } else {
    putdcc $idx "No user flood check data."
  }
  return
}

##
## Display flood checking status
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:floodstatus {hand idx arg} {
  set arg [split [string trimright $arg]]
  set what [lindex $arg 0]
  set ok 1
  switch -exact -- $what {
    all {
      expireFloodData
      FloodStatusSettings $idx
      # putdcc can't handle ""
      putdcc $idx " "
      FloodStatusGlobal $idx
      # putdcc can't handle ""
      putdcc $idx " "
      FloodStatusUser $idx
    }
    settings {
      FloodStatusSettings $idx
    }
    global {
      expireGlobalFloodData
      FloodStatusGlobal $idx
    }
    user {
      expireUserFloodData
      FloodStatusUser $idx
    }
    default {
      putdcc $idx "Usage: floodstatus <option>"
      putdcc $idx "Valid options: all, settings, global, user"
      set ok 0
    }
  }
  if {$ok} then {
    # putdcc can't handle ""
    putdcc $idx " "
    putdcc $idx "End of flood check info."
  }
  return 1
}

##
## Run hourly
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:hourly {minute hour day month year} {
  expireFloodData
  return 0
}

##
## Search command list for a matching command
##
## Args:    command, list
## Returns: index of matching command if one found
##          -1 otherwise
##
proc FindCommand {command list} {
  set index 0
  foreach cmd $list {
    foreach {cmdname cmdcount cmdtime} $cmd {break}
    if {![string compare $cmdname $command]} then {
      return "$index $cmdname $cmdcount $cmdtime"
    }
    incr index
  }
  return -1
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData
  variable FloodConfig

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      gflood {
        scan $configData(gflood) "%\[^:\]:%\[^:\]:" FloodConfig(gfcount) FloodConfig(gftime)
      }
      gthrottle {
        scan $configData(gthrottle) "%\[^:\]:%\[^:\]:" FloodConfig(gtcount) FloodConfig(gttime)
      }
      uflood {
        scan $configData(uflood) "%\[^:\]:%\[^:\]:" FloodConfig(ufcount) FloodConfig(uftime)
      }
      uthrottle {
        scan $configData(uthrottle) "%\[^:\]:%\[^:\]:" FloodConfig(utcount) FloodConfig(uttime)
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData
  variable FloodConfig

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  scan $configData(gflood) "%\[^:\]:%\[^:\]:" FloodConfig(gfcount) FloodConfig(gftime)
  scan $configData(gthrottle) "%\[^:\]:%\[^:\]:" FloodConfig(gtcount) FloodConfig(gttime)
  scan $configData(uflood) "%\[^:\]:%\[^:\]:" FloodConfig(ufcount) FloodConfig(uftime)
  scan $configData(uthrottle) "%\[^:\]:%\[^:\]:" FloodConfig(utcount) FloodConfig(uttime)
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  gflood {group 20:15 {} {count:seconds} {(global) how many commands in number of seconds trigger flood}}
  gfloodtime {integer 120 {} seconds {(global) how long to ignore commands}}
  gthrottle {group 15:10 {} {count:seconds} {(global) how many commands in number of seconds trigger throttle}}
  gthrottime {integer 15 {} seconds {(global) how long to throttle a command}}
  uflood {group 12:15 {} {count:seconds} {(user) how many commands in number of seconds trigger flood}}
  ufloodtime {integer 90 {} seconds {(user) how long to ignore commands}}
  uthrottle {group 3:10 {} {count:seconds} {(user) how many commands in number of seconds trigger throttle}}
  uthrottime {integer 30 {} seconds {(user) how long to throttle a command}}
  ignorecount {integer 5 {} number {add an ignore for someone this many commands after triggering flood}}
  ignoretime {integer 600 {} seconds {seconds to completely ignore someone}}
}

# Bind defaults
array set bindDefaults {
  dcc:floodstatus {{dcc m|- floodstatus {} {}}}
  time:hourly {{time -|- {00 * * * *} {} {}}}
}

# Help Data
array set helpData {
  dcc:floodstatus {{<global|user|all|settings>} {show flood checking status and settings}}
}
