###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: irctools.tcl,v 1.22 2005/01/04 03:36:43 tothwolf Exp $
##

# name: irctools
# version: 1.1
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: irctools.conf
# description: IRC functions for use by other modules.
# provides: putnot putmsg putact putsnd putctcp putctcr putchat put \
#	putWhich putList matchServer formatChanJoin formatChanIdle \
#	checkChannelSetting listChannelsWithSetting commandChar
# requires: listModules moduleBindUnbind slindex


##
## Send out the given text to the given dest as a notice
##
## Args:    dest, text
## Returns: nothing
##
proc putnot {dest text} {
  puthelp "NOTICE $dest :$text"
  return
}

##
## Send out the given text to the given dest as a privmsg
##
## Args:    dest, text
## Returns: nothing
##
proc putmsg {dest text} {
  puthelp "PRIVMSG $dest :$text"
  return
}

##
## Send out the given text to the given dest as an action
##
## Args:    dest, text
## Returns: nothing
##
proc putact {dest text} {
  puthelp "PRIVMSG $dest :\001ACTION $text\001"
  return
}

##
## Send out the given text to the given dest as a sound request
##
## Args:    dest, text
## Returns: nothing
##
proc putsnd {dest text} {
  putserv "PRIVMSG $dest :\001SOUND $text\001"
  return
}

##
## Send out the given text to the given dest as a ctcp request
##
## Args:    dest, text
## Returns: nothing
##
proc putctcp {dest text} {
  putserv "PRIVMSG $dest :\001$text\001"
  return
}

##
## Send out the given text to the given dest as a ctcp reply
##
## Args:    dest, text
## Returns: nothing
##
proc putctcr {dest text} {
  puthelp "NOTICE $dest :\001$text\001"
  return
}

##
## Send out a dcc chat request with a given port
##
## Args:    dest, port
## Returns: nothing
##
proc putchat {dest port} {
  putserv "PRIVMSG $dest :\001DCC CHAT chat [myip] $port\001"
  return
}

##
## Send out the given text to a given dest in a given manner
##
## Args:    action, dest, text
## Returns: nothing
##
proc put {action dest text} {
  $action $dest $text
  return
}

##
## Send out the given text to a given dest in a given manner
##
## Args:    action, dest, text
## Returns: nothing
##
proc putWhich {action dest text {which 0}} {
  if {[string compare "" [set actionSplit [slindex $action $which :]]]} then {
    set action $actionSplit
  }
  if {[string compare "" [set destSplit [slindex $dest $which :]]]} then {
    set dest $destSplit
  }
  $action $dest $text
  return
}

##
## Do something with a list
##
## Args:    action, dest, list, prefix1, prefix2
## Returns: nothing
##
proc putList {action dest list prefix1 prefix2 {which 0}} {
  if {[string compare "" $list]} then {
    if {[string compare "" [set actionSplit [slindex $action $which :]]]} then {
      set action $actionSplit
    }
    if {[string compare "" [set destSplit [slindex $dest $which :]]]} then {
      set dest $destSplit
    }
    # Use prefix1 for index 0 and prefix2 for all the rest
    $action $dest "$prefix1[lindex $list 0]"
    set list [lreplace $list 0 0]
    foreach line $list {
      $action $dest $prefix2$line
    }
  }
  return
}

##
## Return index of closest match in a server list
##
## Args:    list, what, match mode {0 (first ambiguous), 1 (exact match)}
## Returns: index of match in list
##          -1 = no match
##          -2 = ambiguous match if no exact match
##
proc matchServer {list what {mode 1}} {
  set ret -1
  set found 0
  set ambiguous 0
  set listLower [string tolower $list]
  set whatLower [string tolower $what]
  switch -exact -- [llength [split $whatLower :]] {
    3 {
      # search for 'server:port:pass', and only return 3 part match
      set index 0
      foreach server $listLower {
        set tmp [lindex $server 0]
        if {[llength [split $tmp :]] == 3} then {
          if {![string compare $tmp $whatLower]} then {
            set ret $index
          }
        }
        incr index
      }
    }
    2 {
      # search for 'server:port', and return 2 or 3 part match
      set index 0
      foreach server $listLower {
        set tmp [lindex $server 0]
        switch -exact -- [llength [split $tmp :]] {
          2 {
            if {![string compare $tmp $whatLower]} then {
              set ret $index
              set found 1
            }
          }
          3 {
            if {[regexp -- "$whatLower:.*" $tmp]} then {
              if {!$found} then {
                if {!$ambiguous} then {
                  set ret $index
                  set ambiguous 1
                } elseif {!$mode} then {
                  set ret $index
                  set found 1
                } else {
                  set ret -2
                }
              }
            }
          }
        }
        incr index
      }
    }
    1 {
      # search for 'server', and return 1, 2 or 3 part match
      set index 0
      foreach server $listLower {
        set tmp [lindex $server 0]
        switch -exact -- [llength [split $tmp :]] {
          1 {
            if {![string compare $tmp $whatLower]} then {
              set ret $index
              set found 1
            }
          }
          2 {
            if {[regexp -- "$whatLower:.*" $tmp]} then {
              if {!$found} then {
                if {!$ambiguous} then {
                  set ret $index
                  set ambiguous 1
                } elseif {!$mode} then {
                  set ret $index
                  set found 1
                } else {
                  set ret -2
                }
              }
            }
          }
          3 {
            if {[regexp -- "$whatLower:.*:.*" $tmp]} then {
              if {!$found} then {
                if {!$ambiguous} then {
                  set ret $index
                  set ambiguous 1
                } elseif {!$mode} then {
                  set ret $index
                  set found 1
                } else {
                  set ret -2
                }
              }
            }
          }
        }
        incr index
      }
    }
  }
  return $ret
}

##
## Convert join time into '19Dec' and '01:04' formats
##
## Args:    unixtime timestamp
## Returns: date format based on given timestamp
##
proc formatChanJoin {unixtime} {
  if {[expr [clock seconds] - $unixtime] >= 86400} then {
    return [clock format $unixtime -format "%d%b"]
  } else {
    return [clock format $unixtime -format "%H:%M"]
  }
}

##
## Convert idle minute count into '4d', '1h', and '34m' formats
##
## Args:    minutes
## Returns: day, hour, or minute count
##
# returns converted minutes
proc formatChanIdle {minutes} {
  if {$minutes >= 1440} then {
    return [expr $minutes / 1440]d
  } elseif {$minutes >= 60} then {
    return [expr $minutes / 60]h
  }
  return ${minutes}m
}

##
## Check the given channel for the given setting
##
## Args:    channel, setting
## Returns: 1 if the channel has the setting in its info
##          0 otherwise
##
proc checkChannelSetting {chan setting} {
  if {[lsearch -exact [channel info $chan] $setting] != -1} then {
    return 1
  }
  return 0
}

##
## Return a list of channels that have the given setting
##
## Args:    setting
## Returns: list of channels that have the given setting
##
proc listChannelsWithSetting {setting} {
  set list ""
  foreach chan [channels] {
    if {[lsearch -exact [channel info $chan] $setting] != -1} then {
      lappend list $chan
    }
  }
  return $list
}

##
## Return current command character
##
## Args:    type {pub,msg,dcc,fil}
## Returns: command character for the type given
##
proc commandChar {type} {
  global botnick
  variable configData

  switch -exact -- $type {
    pub {
      return $configData(commandchar)
    }
    msg {
      return "/msg $botnick "
    }
    dcc {
      return
    }
    fil {
      return
    }
    default {
      error "Unknown command character type: $type"
    }
  }
  return
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      commandchar {
        foreach module [listModules 1] {
          moduleBindUnbind unbind $module "type pub" "option cmdchr" [list cmdsub "{commandChar pub} ^$oldConfigData(commandchar)"]
          moduleBindUnbind bind $module "type pub" "option cmdchr"
        }
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  commandchar {word - {} character {character to prefix public commands}}
}
