###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: misctools.tcl,v 1.19 2005/01/04 03:36:43 tothwolf Exp $
##

# name: misctools
# version: 0.2
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: misctools.conf
# description: Misc functions for use by other modules
# requires: commandChar putnot contactAddress
# provides: badArgs listBinds countBinds matchBinds findBinds checkKnownUser


##
## Check argument usage for commands
##
## Args:    argument list, min args, max args, command usage example
## Returns: nothing
##
proc badArgs {arg min max example} {
  if {([set length [llength $arg]] < $min) || ($length > $max)} then {
    error "wrong # args: should be \"$example\""
  }
  return
}

##
## List commands that match a particular type and mask
##
## Args:    type, mask
## Returns: list of matching binds
##
proc listBinds {cmptype cmpmask} {
  set result ""
  set binds [binds $cmptype]
  if {[string compare "" $binds]} then {
    foreach bind $binds {
      foreach {type flags name hits proc} $bind {break}
      if {[string match $cmpmask $name]} then {
        lappend result $bind
      }
    }
  }
  return $result
}

##
## Count commands that match a particular type and mask
##
## Args:    type, mask
## Returns: number of matching binds
##
proc countBinds {cmptype cmpmask} {
  set result 0
  set binds [binds $cmptype]
  if {[string compare "" $binds]} then {
    foreach bind $binds {
      foreach {type flags name hits proc} $bind {break}
      if {[string match $cmpmask $name]} then {
        incr result
      }
    }
  }
  return $result
}

##
## Check if the given type and mask matches a bind
##
## Args:    type, mask
## Returns: 1 if the given mask matches a bind
##          0 otherwise
##
proc matchBinds {cmptype cmpmask} {
  set binds [binds $cmptype]
  if {[string compare "" $binds]} then {
    foreach bind $binds {
      foreach {type flags name hits proc} $bind {break}
      if {[string match $cmpmask $name]} then {
        return 1
      }
    }
  }
  return 0
}

##
## List bind that matches the given type and command
##
## Args:    type, command, resultVar
## Returns: 0 if command not found
##          1 if command is exactly matched
##          2 if command is matched
##          -1 if command is ambiguous
##
proc findBinds {cmptype command resultVar} {
  upvar 1 $resultVar result

  set binds [binds $cmptype]
  if {[string compare "" $binds]} then {
    set found 0
    set ambiguous 0
    set lowerCommand [string tolower $command]

    foreach bind $binds {
      foreach {type flags name hits proc} $bind {break}
      # Exact match
      if {![string compare $command $name]} then {
        set result $name
        return 1
      # Partial match
      } elseif {[string match "$lowerCommand*" [string tolower $name]]} then {
        if {!$found} then {
          set result $name
          set found 1
        } else {
          set ambiguous 1
        }
      }
    }
    if {$ambiguous} then {
      return -1
    } elseif {$found} then {
      return 2
    }
  }
  return 0
}

##
## Check the given nick/handle to see if the user is known
##
## Args:    nick, handle
## Returns: 0 if the given handle is not a valid user
##          -1 if the given handle is a bot
##          1 if the given handle is a valid user
##
proc checkKnownUser {nick hand} {
  global learn-users quiet-reject
  variable configData

  if {[validuser $hand]} then {
    if {![matchattr $hand b]} then {
      # User is known
      return 1
    } else {
      # User is known and is a bot
      return -1
    }
  } else {
    # User not known
    if {!${quiet-reject}} then {
      if {${learn-users}} then {
        # Tell the user how to introduce themself
        putnot $nick "I don't know you; please introduce yourself first."
        putnot $nick "[commandChar msg]hello"
      } elseif {($configData(introemail)) && \
                ([string compare "" [set address [contactAddress]]])} then {
        putnot $nick "I don't know you; please email $address or have someone introduce you to me."
      } else {
        putnot $nick "I don't know you; please have someone introduce you to me."
      }
    }
  }
  return 0
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  introemail {switch 0 {} {0/1} {Show contact email address to unknown users?}}
}
