###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: timetools.tcl,v 1.15 2005/01/14 07:22:12 tothwolf Exp $
##

# name: timetools
# version: 1.3
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: timetools.conf
# description: Time formatting functions for use by other modules.
# provides: timerExists utimerExists timerId utimerId killTimerCommand \
#	killUtimerCommand convertTime formatTime daysLeft convertSeconds \
#	convertDHM clockSecondsClicks formatElapsedTime \
#	countClicksPerSecond averageClicksPerSecond \
#	calibrateClicksPerSecond clicksPerSecond
# requires: wpLog


##
## Check if a given command has a timer pending
##
## Args:    matchcmd
## Returns: 1 if the given matchcmd has a pending timer
##          0 otherwise
##
proc timerExists {matchcmd} {
  foreach timer [timers] {
    foreach {seconds command id} $timer {break}
    if {![string compare $matchcmd $command]} then {
      return 1
    }
  }
  return 0
}

##
## Check if a given command has a utimer pending
##
## Args:    matchcmd
## Returns: 1 if the given matchcmd has a pending utimer
##          0 otherwise
##
proc utimerExists {matchcmd} {
  foreach utimer [utimers] {
    foreach {seconds command id} $utimer {break}
    if {![string compare $matchcmd $command]} then {
      return 1
    }
  }
  return 0
}

##
## Find timerid for a given command
##
## Args:    matchcmd
## Returns: timerid if the given command has a pending timer
##          nothing otherwise
##
proc timerId {matchcmd} {
  foreach timer [timers] {
    foreach {seconds command id} $timer {break}
    if {![string compare $matchcmd $command]} then {
      return $id
    }
  }
  return
}

##
## Find timerid for a given command
##
## Args:    matchcmd
## Returns: timerid if the given command has a pending timer
##          nothing otherwise
##
proc utimerId {matchcmd} {
  foreach utimer [utimers] {
    foreach {seconds command id} $utimer {break}
    if {![string compare $matchcmd $command]} then {
      return $id
    }
  }
  return
}

##
## Kill timers matching the given command
##
## Args:    matchcmd
## Returns: number of timers killed
##
proc killTimerCommand {matchcmd} {
  set count 0
  foreach timer [timers] {
    foreach {seconds command id} $timer {break}
    if {![string compare $matchcmd $command]} then {
      killtimer $id
      incr count
    }
  }
  return $count
}

##
## Kill utimers matching the given command
##
## Args:    matchcmd
## Returns: number of utimers killed
##
proc killUtimerCommand {matchcmd} {
  set count 0
  foreach utimer [utimers] {
    foreach {seconds command id} $utimer {break}
    if {![string compare $matchcmd $command]} then {
      killutimer $id
      incr count
    }
  }
  return $count
}

##
## Convert a unixtime timestamp to readable format
##
## Args:    unixtime timestamp
## Returns: 'HH:MM:SS' if less then 1 day ago,
##          'MMM DD YYYY' if more then 1 day ago
##
proc convertTime {time} {
  if {[expr ([clock seconds] - $time) > 86400]} then {
    set reply [clock format $time -format "%b %d %Y"]
  } else {
    set reply [clock format $time -format "%H:%M:%S"]
  }
  return $reply
}

##
## Convert a unixtime timestamp to 'MMM DD YYYY HH:MM:SS' format
##
## Args:    unixtime timestamp
## Returns: 'MMM DD YYYY HH:MM:SS'
##
proc formatTime {time} {
  return [clock format $time -format "%b %d %Y %H:%M:%S"]
}

##
## Calculate days remaining from the unixtime timestamp and day count given
##
## Args:    unixtime timestamp, days
## Returns: days remaining
##
proc daysLeft {time days} {
  set seconds [expr [clock seconds] - $time]
  while {$seconds >= 86400} {
    incr seconds -86400
    incr days -1
  }
  if {$days > 0} then {
    return $days
  }
  return 0
}

##
## Convert a value to/from seconds
##
## Args:    mode {from,to}, what, value
## Returns: converted value
##
proc convertSeconds {mode what value} {
  switch -exact -- $mode {
    from {
      set expr *
    }
    to {
      set expr /
    }
    default {
      return
    }
  }
  switch -exact -- $what {
    seconds {
      return $value
    }
    minutes {
      return [expr $value $expr 60]
    }
    hours {
      return [expr $value $expr 3600]
    }
    days {
      return [expr $value $expr 86400]
    }
    weeks {
      return [expr $value $expr 604800]
    }
    months {
      return [expr $value $expr 2628000]
    }
    years {
      return [expr $value $expr 31536000]
    }
  }
  return
}

##
## Converts XdXhXm time description to number of seconds
##
## Args:    string
## Returns: seconds
##
proc convertDHM {string} {
  set seconds 0
  foreach {type number} [splitNumber $string] {
    switch -exact -- [string tolower $type] {
      d {
        incr seconds [convertSeconds from days $number]
      }
      h {
        incr seconds [convertSeconds from hours $number]
      }
      m {
        incr seconds [convertSeconds from minutes $number]
      }
    }
  }
  return $seconds
}

##
## Create a high resolution timestamp
##
## Arg:  none
## Returns: list of '<seconds> <clicks>'
##
proc clockSecondsClicks {} {
  variable Milliseconds

  if {$Milliseconds} then {
    return "[clock seconds] [clock clicks -milliseconds]"
  }
  return "[clock seconds] [clock clicks]"
}

##
## Format elapsed time
##
## Args:    then {seconds, clicks}, now {seconds, clicks}
## Returns: elapsed time string
##
proc formatElapsedTime {then now} {
  variable Milliseconds

  if {$Milliseconds} then {
    set clicksPerSecond 1000
    set format "%.3f"
  } else {
    set clicksPerSecond [clicksPerSecond]
    set format "%f"
  }

  if {([set time [expr [lindex $now 0] - [lindex $then 0]]] < 60) &&
      ([set clicktime [expr double([lindex $now 1] - [lindex $then 1]) / $clicksPerSecond]] >= 0)} then {
    return "[format $format $clicktime] seconds"
  }
  return [duration $time]
}

##
## Count number of clock clicks in 1 second
##
## Args:    none
## Returns: number of clock clicks in 1 second
##
proc countClicksPerSecond {} {
  set clicks [clock clicks]
  after 1000
  return [expr [clock clicks] - $clicks]
}

##
## Average number of clock clicks per second by number of rounds
##
## Args:    rounds
## Returns: average number of clock clicks
##
proc averageClicksPerSecond {{rounds 5}} {
  set totalClicks 0
  for {set i 0} {$i < $rounds} {incr i} {
    incr totalClicks [countClicksPerSecond]
  }
  return [expr $totalClicks / $rounds]
}

##
## Calibrate number of clock clicks in 1 second
##
## Args:    none
## Returns: nothing
##
proc calibrateClicksPerSecond {} {
  variable configData
  variable clicksPerSecond

  if {![info exists clicksPerSecond]} then {
    wpLog o * "Calibrating time routines -- this will take $configData(calrounds) seconds."
  } else {
    wpLog o * "Recalibrating time routines -- this will take $configData(calrounds) seconds."
  }
  set clicksPerSecond "[averageClicksPerSecond $configData(calrounds)] [clock seconds]"
}

##
## Return number of clock clicks in 1 second
##
## Args:    none
## Returns: clock clicks
##
proc clicksPerSecond {} {
  variable clicksPerSecond

  if {![info exists clicksPerSecond]} then {
    calibrateClicksPerSecond
  }
  return [lindex $clicksPerSecond 0]
}

##
## Recalibrate clock clicks per second if needed
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:hourly {minute hour day month year} {
  variable configData
  variable clicksPerSecond
  variable Milliseconds

  if {(!$Milliseconds) &&
      ((![info exists clicksPerSecond]) ||
       ([expr ([clock seconds] - [lindex $clicksPerSecond 1]) / 3600] >= $configData(recaltime)))} then {
    calibrateClicksPerSecond
  }
  return 0
}

##
## Init module specific things
##
## Args:    none
## Returns: nothing
##
proc ModulePostInit {} {
  variable configData
  variable clicksPerSecond
  variable Milliseconds

  if {[info tclversion] >= 8.3} then {
    set Milliseconds 1
  } else {
    set Milliseconds 0
    if {(![info exists clicksPerSecond]) || \
        ([expr ([clock seconds] - [lindex $clicksPerSecond 1]) / 3600] >= $configData(recaltime))} then {
      calibrateClicksPerSecond
    }
  }
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  calrounds {integer 5 {} rounds {rounds to run when checking clicks per second (at least 5)}}
  recaltime {integer 72 {} hours {hours to wait before recalibrating clicks per second}}
}

# Bind defaults
array set bindDefaults {
  time:hourly {{time -|- {00 * * * *} {} {}}}
}
