###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: userinfo.tcl,v 1.21 2005/01/04 03:36:44 tothwolf Exp $
##

# name: userinfo
# version: 1.3.1
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: userinfo.conf
# description: Provides commands that allow enhancement of `.whois' output \
#	by utilizing the `whois-fields' option of eggdrop.
# requires: replaceExpr authCheck commandChar put wipeFields whoisFields \
#	findUser moduleBindUnbind

#
# Credits and comments for orignal userinfo.tcl are at the end of this file.
#


##
## Set user info: msg
##
## Args:    nick, user@host, handle, arg
## Returns: 1 if user is auth'd
##          0 otherwise
##
proc msg:setuserinfo {nick uhost hand arg} {
  return [setuserinfo msg putnot $nick $nick $uhost $hand "" $arg]
}

##
## Set user info: dcc
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:setuserinfo {hand idx arg} {
  return [setuserinfo dcc putdcc $idx "" "" $hand [lindex [console $idx] 0] $arg]
}

##
## Set user info
##
## Args: type, action, dest, nick user@host, handle, chan, arg
## Returns: 1 if user is auth'd or dcc command is used
##          0 otherwise
##
proc setuserinfo {type action dest nick uhost hand chan arg} {
  global lastbind
  variable configData

  set what $lastbind
  if {[authCheck $type $what $nick $uhost $hand $chan]} then {
    set uwhat [string toupper $what]
    set arg [split [string trimright $arg]]
    if {![string compare "" $arg]} then {
      set info [getuser $hand XTRA $uwhat]
      if {![string compare "" $info]} then {
        put $action $dest "You have no $what set."
      } else {
        put $action $dest "Currently: $info"
      }
    } else {
      set info [replaceExpr [join [lrange $arg 0 159]] "(\\12|\\15)"]
      if {![string compare none [string tolower $info]]} then {
        setuser $hand XTRA $uwhat
        put $action $dest "Removed your $what info."
      } else {
        setuser $hand XTRA $uwhat $info
        put $action $dest "Now: $info"
      }
    }
    return 1
  }
  return 0
}

##
## Change user info for the given user: msg
##
## Args:    nick, user@host, handle, arg
## Returns: 1 if user is auth'd
##          0 otherwise
##
proc msg:chuserinfo {nick uhost hand arg} {
  return [chuserinfo msg putnot $nick $nick $uhost $hand "" $arg]
}

##
## Change user info for the given user: msg
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:chuserinfo {hand idx arg} {
  return [chuserinfo dcc putdcc $idx "" "" $hand [lindex [console $idx] 0] $arg]
}

##
## Change user info for the given user
##
## Args: type, action, dest, nick user@host, handle, chan, arg
## Returns: 1 if user is auth'd or dcc command is used
##          0 otherwise
##
proc chuserinfo {type action dest nick uhost hand chan arg} {
  global lastbind
  variable configData

  set what $lastbind
  if {[authCheck $type $what $nick $uhost $hand $chan]} then {
    set uwhat [string toupper $what]
    set arg [split [string trimright $arg]]
    if {![string compare "" $arg]} then {
      put $action $dest "Usage: [commandChar $type]$what <nick|hand> \[<$what>|none\]"
    } else {
      set who [lindex $arg 0]
      set user [findUser $who $chan]
      if {[validuser $user]} then {
        set info [replaceExpr [join [lrange $arg 1 160]] "(\\12|\\15)"]
        if {(![string compare "" $info]) || \
            (![string compare none [string tolower $info]])} then {
          setuser $user XTRA $uwhat
          put $action $dest "Wiped default $what for $user"
        } else {
          setuser $user XTRA $uwhat $info
          put $action $dest "New default $what for $user: $info"
        }
      } else {
        put $action $dest "I don't know who $who is."
      }
    }
    return 1
  }
  return 0
}

##
## Show all userinfo fields: dcc
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:showfields {hand idx arg} {
  variable configData

  if {![string compare "" $configData(fields)]} then {
    putdcc $idx "There are no userinfo fields defined."
  } else {
    putdcc $idx "Userinfo fields: $configData(fields)"
  }
  return 1
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable ModuleName
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      fields {
        foreach field $oldConfigData(fields) {
          if {[lsearch -exact $configData(fields) $field] == -1} then {
            lappend RemovedFields $field
          }
        }
        if {[info exists RemovedFields]} then {
          wipeFields $RemovedFields
        }
        whoisFields update $configData(fields) $oldConfigData(fields)
        foreach field $oldConfigData(fields) {
          moduleBindUnbind unbind $ModuleName "option noauto" [list regsub "%field% $field"]
        }
        foreach field $configData(fields) {
          moduleBindUnbind bind $ModuleName "option noauto" [list regsub "%field% $field"]
        }
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable ModuleName
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  whoisFields update $configData(fields) $oldConfigData(fields)
  trace variable configData w [namespace code TraceConfigData]
  foreach field $configData(fields) {
    moduleBindUnbind bind $ModuleName "option noauto" [list regsub "%field% $field"]
  }
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable ModuleName
  variable configData

  whoisFields remove $configData(fields)
  trace vdelete configData w [namespace code TraceConfigData]
  foreach field $configData(fields) {
    moduleBindUnbind unbind $ModuleName "option noauto" [list regsub "%field% $field"]
  }
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  fields {string {url bf gf irl email dob} {} fields {extra userinfo fields for users}}
}

# Bind defaults
array set bindDefaults {
  msg:setuserinfo {{msg -|- %field% noauto {}}}
  msg:chuserinfo {{msg m|- ch%field% noauto {}}}
  dcc:setuserinfo {{dcc -|- %field% noauto {}}}
  dcc:chuserinfo {{dcc m|- ch%field% noauto {}}}
  dcc:showfields {{dcc m|- showfields {} {}}}
}

# Help Data
array set helpData {
  msg:setuserinfo {{%field% [<what>|none]} {set user info}}
  msg:chuserinfo {{%field% <nick|hand> [<%field%>|none]} {change user info for the given user}}
  dcc:setuserinfo {{%field% [<what>|none]} {set user info}}
  dcc:chuserinfo {{%field% <nick|hand> [<%field%>|none]} {change user info for the given user}}
  dcc:showfields {{} {show all userinfo fields}}
}


# userinfo.tcl v1.05 for Eggdrop 1.1.6 and higher
#           Scott G. Taylor -- ButchBub!staylor@mrynet.com
#
# v1.00      ButchBub     14 July      1997 -Original release.  Based on
#                                            whois.tcl "URL" commands.
# v1.01      Beldin       11 November  1997 -1.3 only version
# v1.02      Kirk         19 June      1998 -extremely small fixes
# v1.03      guppy        17 March     1999 -small fixes again
# v1.04      Ernst        15 June      1999 -fix for egg 1.3.x + TCL 8.0
# v1.05      Dude         14 July      1999 -small cosmetic/typo fixes
#                                           -$lastbind bug work around fix
#                                           -added userinfo_loaded var
#                                           -fixed bug in dcc_chuserinfo proc
#                                           -unbinds removed user fields
#                                           -dcc .showfields command added
#                                           -deletes removed userinfo fields
#                                            from the whois-fields list.
#
# TO USE:  o    Set the desired userinfo field keywords to the
#               `userinfo-fields' line below where indicated.
#          o    Load this script on a 1.1.6 or later Eggdrop bot.
#          o    Begin having users save the desired information.  If you
#               choose to add the default "IRL" field, they just use
#               the IRC command: /MSG <botnick> irl Joe Blow.
#          o    See the new information now appear with the whois command.
#
# This script enhances the `whois' output utilising the `whois-fields'
# option of eggdrop 1.1-grant and later versions.  It adds the functionality
# of whois.tcl used in pre-1.1-grant versions.
#
# The fields desired to be maintained in the userfile `xtra' information
# should be put in `userinfo-fields'.  This is different than the Eggdrop
# configuration variable `whois-fields' in that this script will add the
# commands to change these fields.   It will also add these desired fields
# to the `whois-fields' itself, so do not define them there as well.  The
# fields added in `userinfo-fields' will be converted to upper case for
# aesthetics in the `whois' command output.
#
# The commands that will be added to the running eggdrop are:
#  (<info> will be the respective userfile field added in `userinfo-fields')
#
#   TYPE    COMMAND         USAGE
#   ======  ==============  ========================================
#   msg      <info>         To change your <info> via /MSG.
#   dcc     .<info>         To change your <info> via DCC.
#   dcc     .ch<info>       To change someone else's <info> via DCC.
#
# Currently supported fields and commands:
#
#   FIELD   USAGE
#   =====   =====================
#   URL     WWW page URL
#   IRL     In Real Life name
#   BF      Boyfriend
#   GF      Girlfriend
#   DOB     Birthday (Date Of Birth)
#   EMAIL   Email address
