###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: userping.tcl,v 1.36 2005/01/04 03:36:44 tothwolf Exp $
##

# name: userping
# version: 1.5.1
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: userping.conf
# description: Adds dcc, msg, and public channel commands called `ping' \
#	that allow the bot to ping users and display the response.
# requires: putctcp putmsg putnot put isNumber clockSecondsClicks \
#	formatElapsedTime authCheck floodCheck commandChar \
#	checkKnownUser grammarA grammarAnd breakDownFlags findUser \
#	findNickOnChannels


##
## Send a ping request to someone: pub
##
## Args:    nick, user@host, handle, channel, arg
## Returns: 1 if the user is auth'd
##          0 otherwise
##
proc pub:ping {nick uhost hand chan arg} {
  variable configData

  if {(![isbotnick $nick]) && \
      ([floodCheck $nick $uhost ping]) && \
      ((!$configData(authmode)) || \
       (($configData(authmode) == 1) && \
        ([checkKnownUser $nick $hand] == 1)) || \
       (($configData(authmode) == 2) && \
        ([authCheck pub ping $nick $uhost $hand $chan])))} then {
    set arg [split [string trimright $arg]]
    if {[string compare "" $arg]} then {
      set who [lindex $arg 0]
    } else {
      set who $nick
    }
    if {[onchan $who $chan]} then {
      if {$configData(pubreply)} then {
        SendCtcpPing $who putmsg $chan
      } else {
        SendCtcpPing $who putnot $nick
      }
    } else {
      putnot $nick "$who isn't on $chan."
    }
    return 1
  }
  return 0
}

##
## Send a ping request to someone: msg
##
## Args:    nick, user@host, handle, arg
## Returns: 1 if the user is auth'd
##          0 otherwise
##
proc msg:ping {nick uhost hand arg} {
  variable configData

  if {(![isbotnick $nick]) && \
      ([floodCheck $nick $uhost ping]) && \
      ([authCheck msg ping $nick $uhost $hand ""])} then {
    set who [lindex [split [string trimright $arg]] 0]
    if {([string compare "" $who]) && \
        ([string compare [string tolower $who] [string tolower $nick]])} then {
      if {([string compare "" $configData(pingflags)]) && \
          (![matchattr $hand $configData(pingflags)])} then {
        putnot $nick "Sorry, only [grammarA [grammarAnd [breakDownFlags $configData(pingflags)] or]] can ping other users."
      } else {
        SendCtcpPing $who putnot $nick
      }
    } else {
      SendCtcpPing $nick putnot $nick
    }
    return 1
  }
  return 0
}

##
## Send a ping request to someone: dcc
##
## Args:    handle, idx, arg
## Returns: 1
##
proc dcc:ping {hand idx arg} {
  variable configData

  set who [lindex [split [string trimright $arg]] 0]
  set nick [findNickOnChannels $hand $idx]
  if {[string compare "" $who]} then {

    # Self ping
    if {([string compare "" $nick]) && \
        ([string compare [string tolower $who] [string tolower $nick]])} then {
      SendCtcpPing $who putdcc $idx

    # User doesn't have access to ping others
    } elseif {([string compare "" $configData(pingflags)]) && \
              (![matchattr $hand $configData(pingflags)])} then {
      putdcc $idx "Sorry, only [grammarA [grammarAnd [breakDownFlags $configData(pingflags)] or]] can ping other users."

    # User pings someone, not really any sane way to test $who here
    } else {
      SendCtcpPing $who putdcc $idx
    }
  } else {

    # Self ping
    if {[string compare "" $nick]} then {
      SendCtcpPing $nick putdcc $idx
    } else {
      putdcc $idx "Sorry, I can't find your nick on any channels that I monitor."
    }
  }
  return 1
}

##
## Display ping reply from a user
##
## Args:    nick, user@host, handle, dest, keyword, arg
## Returns: 0 if the ctcp reply was processed
##          1 otherwise
##
proc ctcr:ping {nick uhost hand dest key arg} {
  variable configData
  variable PingData

  set now [clockSecondsClicks]
  set who [string tolower $nick]
  if {[info exists PingData($who)]} then {
    set arg [split $arg]
    set seconds [lindex $arg 0]
    set clicks [lindex $arg 1]
    if {([isNumber $seconds]) && ([isNumber $clicks])} then {
      put [lindex $PingData($who) 0] [lindex $PingData($who) 1] "Ping reply from $nick: [formatElapsedTime "$seconds $clicks" $now]"
    }
    unset PingData($who)
    return 0
  }
  return 1
}

##
## Clear invalid nicks from PingData array
##
## Args:    from, key, arg
## Returns: 0
##
proc raw:401 {from key arg} {
  variable PingData

  set nick [string tolower [lindex [split $arg] 1]]
  if {[info exists PingData($nick)]} then {
    unset PingData($nick)
  }
  return 0
}

##
## Clear old data from array once an hour
##
## Args:    minute, hour, day, month, year
## Returns: 0
##
proc time:hourly {minute hour day month year} {
  variable PingData

  if {[string compare "" [set pingdata [array get PingData]]]} then {
    set now [clock seconds]
    foreach {name data} $pingdata {
      if {[expr $now - [lindex $data 2]] >= 3600} then {
        unset PingData($name)
      }
    }
  }
  return 0
}

##
## Send a CTCP PING request
##
## Args:    who, action, dest
## Returns: nothing
##
proc SendCtcpPing {who action dest} {
  variable PingData

  set now [clockSecondsClicks]
  set PingData([string tolower $who]) "$action $dest [lindex $now 0]"
  putctcp $who "PING $now"
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  pubreply {switch 1 {} {0/1} {show ping replys on the channel for public ping command?}}
  authmode {{switch {range {0 2}}} 0 {} {0/1/2} {0 - allow anyone to use ping, 1 - only regonized users, 2 - only auth'd users}}
  pingflags {word o|o {} {flag(s)} {flag(s) required to ping someone else}}
}

# Bind defaults
array set bindDefaults {
  pub:ping {{pub -|- ping {cmdchr {cmdsub {{commandChar pub} ^}}} {}}}
  msg:ping {{msg -|- ping {} {}}}
  dcc:ping {{dcc -|- ircping {} {}}}
  ctcr:ping {{ctcr -|- PING {} {}}}
  raw:401 {{raw -|- 401 {} {}}}
  time:hourly {{time -|- {00 * * * *} {} {}}}
}

# Help Data
array set helpData {
  pub:ping {{[nick]} {make the bot ping the given nick, or the issuer of the command if no nick given.}}
  msg:ping {{[nick]} {make the bot ping the given nick, or the issuer of the command if no nick given.}}
  dcc:ping {{[nick]} {make the bot ping the given nick, or the issuer of the command if no nick given.}}
}
