###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: usertools.tcl,v 1.33 2005/01/04 03:36:44 tothwolf Exp $
##

# name: usertools
# version: 1.1
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# description: User and user data functions for use by other modules.
# provides: validHandle checkUser findUser findHandOnChannels \
#	findNickOnChannels findField clearFields wipeFields moveFields \
#	whoisFields getLastOn findChanFlag isBotHand chanFlag \
#	breakDownFlags listUsersWithoutPass searchUserHandles \
#	listUserHandles searchUserHosts listUserHosts searchUsersAndHosts \
#	listUsersAndHosts checkOwner dccCheckOwner
# requires: isNumber breakDownSubString


##
## Check if a handle is valid and is not a bot
##
## Args:    handle
## Returns: 1 if the given handle is valid and is not a bot
##          0 otherwise
##
proc validHandle {hand} {
  if {([validuser $hand]) && (![matchattr $hand b])} then {
    return 1
  }
  return 0
}

##
## Sanity check the given handle
##
## Args:    handle, args {flags, channel}
## Returns: 0 if the given handle is not a valid user
##          -1 if the given handle is a bot
##          -2 if the given handle does not have the given flags
##          1 otherwise
##
proc checkUser {hand args} {
  if {[validuser $hand]} then {
    if {[matchattr $hand b]} then {
      return -1
    } elseif {[string compare "" [set flags [lindex $args 0]]]} then {
      if {([string compare "" [set chan [lindex $args 1]]]) && \
          (![matchattr $hand $flags $chan])} then {
        return -2
      } elseif {![matchattr $hand $flags]} then {
        return -2
      } else {
        return 1
      }
    } else {
      return 1
    }
  }
  return 0
}

##
## Find the handle of the given nick on a channel
##
## Args:    nick/handle, channel
## Returns: handle of the given nick
##
proc findUser {who chan} {
  if {[validuser $who]} then {
    return [finduser [lindex [getuser $who HOSTS] 0]]
  } elseif {([validchan $chan]) && ([onchan $who $chan])} then {
    return [nick2hand $who $chan]
  } else {
    return
  }
}

##
## Search for the given handle on all channels the bot monitors
##
## Args:    handle
## Returns: 1 if the given handle is on any channels the bot monitors
##          0 otherwise
##
proc findHandOnChannels {hand} {
  # NOTE: This proc is useless for part and quit bindings.
  #       Eggdrop's (broken?) behavior leaves user data in the cache
  #       until after the bound procs have been processed.
  foreach chan [channels] {
    if {[handonchan $hand $chan]} then {
      return 1
    }
  }
  return 0
}

##
## Return a nickname for the given handle
##
## Args:   hand, idx
## Returns handle's nickname if found
##         nothing otherwise
##
proc findNickOnChannels {hand {idx ""}} {
  if {[valididx $idx]} then {
    set chan [lindex [console $idx] 0]
    if {([validchan $chan]) && \
        ([handonchan $hand $chan])} then {
      return [hand2nick $hand $chan]
    }
  }
  foreach chan [channels] {
    if {[handonchan $hand $chan]} then {
      return [hand2nick $hand $conchan]
    }
  }
  return
}

##
## Returns list of all users who have data in the given field
##
## Args:    field, data
## Returns: list of users
##
proc findField {field {data ""}} {
  set users ""
  if {![string compare "" $data]} then {
    foreach user [lsort [split [userlist]]] {
      if {[string compare "" [getuser $user XTRA $field]]} then {
        lappend users $user
      }
    }
  } else {
    foreach user [lsort [split [userlist]]] {
      if {([string compare "" [set tmp [getuser $user XTRA $field]]]) && \
          (![string compare $tmp $data])} then {
        lappend users $user
      }
    }
  }
  return $users
}

##
## Clear data for all the given fields for the given handle
##
## Args:    handle, fields
## Returns: nothing
##
proc clearFields {hand fields} {
  foreach field $fields {
    if {[string compare "" [getuser $hand XTRA $field]]} then {
      setuser $hand XTRA $field
    }
  }
  return
}

##
## Remove the given fields from all users
##
## Args:    fields
## Returns: nothing
##
proc wipeFields {fields} {
  foreach user [userlist] {
    foreach field $fields {
      if {[string compare "" [getuser $user XTRA $field]]} then {
        setuser $user XTRA $field
      }
    }
  }
  return
}

##
## Move data from one field to another
##
## Args:    old field, new field
## Returns: nothing
##
proc moveFields {oldfield newfield} {
  foreach user [userlist] {
    if {[string compare "" [set data [getuser $user XTRA $oldfield]]]} then {
      setuser $user XTRA $newfield $data
      setuser $user XTRA $oldfield
    }
  }
  return
}

##
## Add/remove/update fields in eggdrop's whois-fields variable
##
## Args:    action {add|remove|update}
## Returns: 1 if successful
##          0 otherwise
##
proc whoisFields {action fields {oldFields ""}} {
  global whois-fields

  if {![info exists whois-fields]} then {
    set whois-fields ""
  }
  switch -exact -- $action {
    add -
    update {
      # Remove old fields from whois-fields
      if {[string compare "" $oldFields]} then {
        foreach oldField $oldFields {
          if {[lsearch -exact $fields $oldField] == -1} then {
            if {[set index [lsearch -exact ${whois-fields} $oldField]] != -1} then {
              set whois-fields [lreplace ${whois-fields} $index $index]
            }
          }
        }
      }
      # Add fields to whois-fields
      foreach field $fields {
        if {[lsearch -exact ${whois-fields} $field] == -1} then {
          lappend whois-fields $field
        }
      }
    }
    remove {
      # Remove fields from whois-fields
      foreach field $fields {
        if {[set index [lsearch -exact ${whois-fields} $field]] != -1} then {
          set whois-fields [lreplace ${whois-fields} $index $index]
        }
      }
    }
  }
  return
}

##
## Get laston info for a user
##
## Args:    handle, args {channel}
## Returns: readable laston format for a user
##
proc getLastOn {hand args} {
  set chan [lindex $args 0]
  if {![string compare "" $chan]} then {
    set laston [getuser $hand LASTON]
  } else {
    set laston [getuser $hand LASTON $chan]
  }
  set lastutime [lindex $laston 0]
  if {(![string compare "" $lastutime]) || \
      ($lastutime <= 0)} then {
    return "never nowhere"
  } else {
    if {[expr [clock seconds] - $lastutime] > 86400} then {
      set lasttime [clock format $lastutime -format "%d %b"]
    } else {
      set lasttime [clock format $lastutime -format "%H:%M"]
    }
    set lastchan [lindex $laston 1]
    if {![string compare "" $lastchan]} then {
      return [list $lasttime]
    } else {
      return [list $lasttime $lastchan]
    }
  }
}

##
## Search all channel records for the given user for the given flag
##
## Args:   user, flag, channels
## Returns: 1 if the given user has the given flag for any channel
##          0 otherwise
##
proc findChanFlag {user flag {channels ""}} {
  if {![string compare "" $channels]} then {
    set channels [channels]
  }
  foreach chan $channels {
    if {[matchattr $user |$flag $chan]} then {
      return 1
    }
  }
  return 0
}

##
## Compare the given handle to the bot's handle
##
## Args:    hand
## Returns: 1 if hand matches bot's handle
##          0 otherwise
##
proc isBotHand {hand} {
  global botnet-nick

  if {![string compare [string toupper $hand] [string toupper ${botnet-nick}]]} then {
    return 1
  }
  return 0
}

##
## Find the most significant flag for the given user and channel
##
## Args:    user, channel
## Returns: most significant flag for the given user and channel
##
proc chanFlag {user {chan ""}} {
  if {[validchan $chan]} then {
    set flags [chattr $user $chan]
  } else {
    set flags [chattr $user]
  }
  switch -glob -- $flags {
    "*b*\|*" {
      return b
    }
    "*n*\|*" {
      return N
    }
    "*\|*n*" {
      return n
    }
    "*m*\|*" {
      return M
    }
    "*\|*m*" {
      return m
    }
    "*d*\|*" {
      return D
    }
    "*\|*d*" {
      return d
    }
    "*a*\|*" {
      return A
    }
    "*\|*a*" {
      return a
    }
    "*o*\|*" {
      return O
    }
    "*\|*o*" {
      return o
    }
    "*q*\|*" {
      return Q
    }
    "*\|*q*" {
      return q
    }
    "*g*\|*" {
      return G
    }
    "*\|*g*" {
      return g
    }
    "*v*\|*" {
      return V
    }
    "*\|*v*" {
      return v
    }
  }
  return " "
}

##
## Break down a list of flags and return a description for each flag
##
## Args:    flags, type
## Returns: descriptive list of flags
##
proc breakDownFlags {flags {type ""}} {

  array set flagTypes {
    global globalFlags
    channel channelFlags
    bot botFlags
    botchannel botChannelFlags
  }

  array set globalFlags {
    - any
    a autoop
    b bot
    c common
    d deop
    e {nethack exempt}
    f friend
    g autovoice
    h highlight
    j janitor
    k autokick
    l halfop
    m master
    n owner
    o op
    p {party line}
    q quiet
    r dehalfop
    t {botnet master}
    u unshared
    v voice
    w {wasop test}
    x xfer
    y autohalfop
    z {washalfop test}
  }

  array set channelFlags {
    - any
    n owner
    m master
    o op
    l halfop
    v voice
    f friend
    a autoop
    y autohalfop
    g autovoice
    d deop
    r dehalfop
    q quiet
    k autokick
    e {nethack exempt}
    w {wasop test}
    z {washalfop test}
  }

  array set botFlags {
    - any
    h {primary hub}
    a {alternate hub}
    l leaf
    r reject
    i isolate
    s {passive share}
    p {agressive share}
    g {global share}
  }

  array set botChannelFlags {
    - any
    s {passive share}
  }

  if {![string compare "" $type]} then {
    set type global
  }

  if {[info exists flagTypes($type)]} then {
    set array $flagTypes($type)
  }

  # FIXME add support for |, &, ||, &&
  set flags [lindex [split $flags |&] 0]

  if {[info exists array]} then {
    set result [breakDownSubString $array $flags "" 0 ", "]
  }

  if {([info exists result]) && ([string compare "" $result])} then {
    return $result
  }

  return $flags
}

##
## Return a list of users who have no password set
##
## Args:    flags
## Returns: list of users with no password set
##
proc listUsersWithoutPass {{flags ""}} {
  set list ""
  foreach hand [split [userlist $flags]] {
    if {[passwdok $hand ""]} then {
      lappend list $hand
    }
  }
  return $list
}

##
## Search for the given mask in the userlist
##
## Args:    mask
## Returns: 1 if the given mask matched a user in the userlist
##          0 otherwise
##
proc searchUserHandles {mask} {
  set lmask [string tolower $mask]
  foreach user [string tolower [userlist]] {
    if {[string match $lmask $user]} then {
      return 1
    }
  }
  return 0
}

##
## Search for the given mask in the userlist
##
## Args:    mask
## Returns: list of users from the userlist who matched the given mask
##
proc listUserHandles {mask} {
  set list ""
  set lmask [string tolower $mask]
  foreach user [userlist] {
    if {[string match $lmask [string tolower $user]]} then {
      lappend list $user
    }
  }
  return $list
}

##
## Search for the given user's hosts for the given mask
##
## Args:    user, mask
## Returns: 1 if the given mask matched a user's host
##          0 otherwise
##
proc searchUserHosts {user mask} {
  set lmask [string tolower $mask]
  foreach host [string tolower [getuser $user HOSTS]] {
    if {[string match $lmask $host]} then {
      return 1
    }
  }
  return 0
}

##
## Search for the given user's hosts for the given mask
##
## Args:    user, mask
## Returns: list of user's hosts that matched the given mask
##
proc listUserHosts {user mask} {
  set list ""
  set lmask [string tolower $mask]
  foreach host [getuser $user HOSTS] {
    if {[string match $lmask [string tolower $host]]} then {
      lappend list $host
    }
  }
  return $list
}

##
## Search the userlist for a handle or host that matches the given mask
##
## Args:    mask
## Returns: 1 if the given mask matched a handle or host in the userlist
##          0 otherwise
##
proc searchUsersAndHosts {mask} {
  set lmask [string tolower $mask]
  foreach user [string tolower [userlist]] {
    if {([string match $lmask $user]) || \
        ([searchUserHosts $user $mask])} then {
      return 1
    }
  }
  return 0
}

##
## Search the userlist for a handle or host that matches the given mask
##
## Args:    mask
## Returns: list of handles that matched the given mask
##
proc listUsersAndHosts {mask} {
  set list ""
  set lmask [string tolower $mask]
  foreach user [userlist] {
    if {([string match $lmask [string tolower $user]]) || \
        ([searchUserHosts $user $mask])} then {
      lappend list $user
    }
  }
  return $list
}

##
## Check if the given handle is a (permanent) owner
##
## Args:    handle, level
## Returns: if must-be-owner is greater than or equal to the given level,
##            1 if given handle is a permanent owner
##            -1 if the given handle is an owner
##          else,
##            1 if the given handle is an owner
##          0 otherwise
##
proc checkOwner {hand {level 1}} {
  global owner must-be-owner

  if {[matchattr $hand n]} then {
    if {([isNumber ${must-be-owner}]) && (${must-be-owner} >= $level)} then {
      regsub -all -- {,} $owner {} ownerList
      if {[lsearch -exact [string tolower $ownerList] [string tolower $hand]] != -1} then {
        return 1
      } else {
        return -1
      }
    } else {
      return 1
    }
  }
  return 0
}

##
## Check if the given handle is an owner
##
## Args:    handle, idx, level
## Returns: 1 if the given handle is an owner
##          0 otherwise
##
proc dccCheckOwner {hand idx {level 1}} {
# FIXME: add support for 'HQ' user for -nt operation
  if {[checkOwner $hand $level] != 1} then {
    putdcc $idx "What?  You need '.help'"
    return 0
  }
  return 1
}
