###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: fooddata.tcl,v 1.12 2005/01/04 03:36:44 tothwolf Exp $
##

# name: wrpg-fooddata
# version: 1.0
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: wrpg-fooddata.conf
# description:
# requires: dataFormatValue dataFormatReplace daysLeft arrayFindElementName
# provides: foodData


# FIXME: finish cleanup, test and debug

##
## Handle food data
##
## Args:    action {add|del|set|get|list|exists|realname|format|value}, args
## Returns: data passed from sub procs
##
proc foodData {action args} {
  switch -exact -- $action {
    add {
      return [foodDataAdd $args]
    }
    del {
      return [foodDataDel $args]
    }
    set {
      return [foodDataSet $args]
    }
    get {
      return [foodDataGet $args]
    }
    list {
      return [foodDataList $args]
    }
    exists {
      return [foodDataExists $args]
    }
    realname {
      return [foodDataRealname $args]
    }
    format {
      return [foodDataFormat $args]
    }
    value {
  	return [foodDataValue $args]
    }
  }
  return
}

##
## Add food data to the food list
##
## Args:    args {name, amount}
## Returns: {1 realname} if name added to food list
##          0 if name is not in hunt list
##          {-1 realname} if amount is invalid
##
proc foodDataAdd {args} {
  variable dataFormats
  variable fooddata

  set realname [huntData realname [lindex $args 0]]
  if {[string compare "" $realname]} then {
    set amount [lindex $args 1]
    if {![string compare "" $amount]} then {
      set amount [huntData value $realname food]
    } elseif {(![isNumber $amount]) || ($amount < 1)} then {
      return [list -1 $realname]
    }
    if {[info exists fooddata($realname)]} then {
      set oldamount [dataFormatValue $dataFormats(fooddata) $fooddata($realname) amount]
      set fooddata($realname) [dataFormatReplace $dataFormats(fooddata) $fooddata($realname) amount [expr $oldamount + $amount]]
      return [list 1 $realname]
    } else {
      set timestamp [clock seconds]
      set fooddata($realname) [dataFormatReplace $dataFormats(fooddata) $fooddata($realname) "amount timestamp" "$amount [clock seconds]"]
      return [list 1 $realname]
    }
  }
  return 0
}

##
## Remove food data from the food list
##
## Args:    args {name, amount}
## Returns: {2 realname} if name removed from food list
##          {1 realname} if amount removed from food list
##          0 if name is not in hunt list
##          {-1 realname} if amount is invalid
##          {-2 realname} if name is not in food list
##
proc foodDataDel {args} {
  variable dataFormats
  variable fooddata

  set realname [huntData realname [lindex $args 0]]
  if {[string compare "" $realname]} then {
    if {[info exists fooddata($realname)]} then {
      set amount [lindex $args 1]
      set oldamount [dataFormatValue $dataFormats(fooddata) $fooddata($realname) amount]
      if {(![string compare "" $amount]) || \
          (![string compare $amount $oldamount])} then {
        unset fooddata($realname)
        return [list 2 $realname]
      } elseif {([isNumber $amount]) && \
                ($amount >= 1) && \
                ($amount <= $oldamount)} then {
        set fooddata($realname) [dataFormatReplace $dataFormats(fooddata) $fooddata($realname) amount [expr $oldamount - $amount]]
        return [list 1 $realname]
      } else {
        return [list -1 $realname]
      }
    } else {
      return [list -2 $realname]
    }
  }
  return 0
}

##
## Set food data amount in the food list
##
## Args:    args {name, amount}
## Returns: {1 realname} if name set to amount
##          0 if name is not in hunt list
##          {-1 realname} if amount is invalid
##          {-2 realname} if name is not in food list
##
proc foodDataSet {args} {
  variable fooddata

  set realname [huntData realname [lindex $args 0]]
  if {[string compare "" $realname]} then {
    if {[info exists fooddata($realname)]} then {
      set amount [lindex $args 1]
      if {([isNumber $amount]) && ($amount >= 0)} then {
        set fooddata($realname) [dataFormatReplace $dataFormats(fooddata) $fooddata($realname) amount $amount]
        return [list 1 $realname]
      } else {
        return [list -1 $realname]
      }
    } else {
      return [list -2 $realname]
    }
  }
  return 0
}

##
## Get food data amount from the food list
##
## Args:    args {name}
## Returns: food data amount for 'name'
##
proc foodDataGet {args} {
  variable fooddata

  set realname [huntData realname [lindex $args 0]]
  if {([string compare "" $realname]) && \
      ([info exists fooddata($realname)])} then {
    return $fooddata($realname)
  }
  return
}

##
## List food data from the food list
##
## Args:    args {what}
## Returns: {3 data} if what is a valid name in food list
##          {2 count list} if what is a valid day count in food list
##          {1 count list} if what is not specified
##          0 if name is not in hunt list
##          -1 if food list is empty
##          -2 if no food data has the specified day count
##          -3 if food never spoils
##          {-4 realname} if name is not in food list
##
proc foodDataList {args} {
  variable configData
  variable dataFormats
  variable fooddata

  set what [lindex $args 0]
  set count 0
  if {![string compare "" $what]} then {
    foreach name [lsort [array names fooddata]] {
      set amount [dataFormatValue $dataFormats(fooddata) $fooddata($name) amount]
      set timestamp [dataFormatValue $dataFormats(fooddata) $fooddata($name) timestamp]
      if {$configData(foodspoil)} then {
        lappend list "$name \[$amount:[daysLeft $timestamp $configData(foodspoil)]\],"
      } else {
        lappend list "$name \[$amount\],"
      }
      incr count
    }
    if {$count} then {
      return [list 1 $count $list]
    } else {
      return -1
    }
  } elseif {[isNumber $what]} then {
    if {$configData(foodspoil)} then {
      foreach name [lsort [array names fooddata]] {
        set amount [dataFormatValue $dataFormats(fooddata) $fooddata($name) amount]
        set timestamp [dataFormatValue $dataFormats(fooddata) $fooddata($name) timestamp]
        set age [daysLeft $timestamp $configData(foodspoil)]
        if {![string compare $what $age]} then {
          lappend list "$name \[$amount:$age\],"
          incr count
        }
      }
      if {$count} then {
        return [list 2 $count $list]
      } else {
        return -2
      }
    } else {
      return -3
    }
  } else {
    set realname [huntData realname $what]
    if {[string compare "" $realname]} then {
      if {[info exists fooddata($realname)]} then {
        set amount [dataFormatValue $dataFormats(fooddata) $fooddata($realname) amount]
        set timestamp [dataFormatValue $dataFormats(fooddata) $fooddata($realname) timestamp]
        if {$configData(foodspoil)} then {
          return [list 3 "$realname \[$amount:[daysLeft $timestamp $configData(foodspoil)]\]"]
        } else {
          return [list 3 "$name \[$amount\]"]
        }
      } else {
        return [list -4 $realname]
      }
    }
  }
  return 0
}

##
## Check if food data exists in the food list
##
## Args:    args {name}
## Returns: 1 if name is in the food list
##          0 if name is not in the food list or hunt list
##
proc foodDataExists {args} {
  variable fooddata

  set realname [huntData realname [lindex $args 0]]
  if {([string compare "" $realname]) && \
      ([info exists fooddata($realname)])} then {
    return 1
  }
  return 0
}

##
## Find realname for food data in the food list
##
## Args:    args {name}
## Returns: realname if name is in the food list
##          "" if name is not in the food list
##
proc foodDataRealname {args} {
  variable fooddata

  set realname [arrayFindElementName fooddata [lindex $args 0]]]
  if {[string compare "" $realname]} then {
    return $realname
  }
  return
}

##
## Return fooddata data format
##
## Args:    args
## Returns: fooddata data format
##
proc foodDataFormat {args} {
  variable dataFormats

  return $dataFormats(fooddata)
}

##
## Return data for a particular fooddata field
##
## Args:    args
## Returns: huntdata data format
##
proc foodDataValue {args} {
  variable fooddata
  variable dataFormats

  set realname [huntData realname [lindex $args 0]]
  set what [lindex $args 1]
  if {([string compare "" $realname]) && \
      ([info exists fooddata($realname)])} then {
    return [dataFormatValue $dataFormats(fooddata) $fooddata($realname) $what]
  }
  return
}

# Set data format information
array set dataFormats {
  fooddata "{amount 1} {timestamp 0}"
}

# Set data file information
array set dataOptions {
  fooddata "array wrpg-food.data food"
}

# Configuration defaults, values and descriptions
array set configDefaults {
  foodspoil {integer 10 {} days {how many days should pass before food spoils}}
  allownonpack {switch 1 {} {0/1} {allow non pack to participate in a hunt?}}
  autocaught {switch 0 {} {0/1} {automatically register a catch?}}
  foodbury {switch 0 {} {0/1} {remove animals from the food list automatically?}}
  timeout {integer 10 {} minutes {minutes before a hunt times out}}
}
