###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: pointdata.tcl,v 1.11 2005/01/04 03:36:44 tothwolf Exp $
##

# name: wrpg-pointdata
# version: 1.0
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: wrpg-pointdata.conf
# description: Wolf RPG user hit point and food point data support.
# requires: isNumber convertTime clearFields whoisFields convertSeconds \
#	userStatusData validHandle
# provides: userPointData updatePointData


##
## Handle a user's hp/fp data
##
## Args:    user, what {hp|fp}, action {incr|decr|set|list|itime|dtime|
##          check|update|test1|test2|limit|clear}, amount
## Returns: hp/fp, time, 0 or -1
##
proc userPointData {user action what args} {
  variable configData

  set amount [lindex $args 0]
  set uwhat [string toupper $what]
  switch -exact -- $action {
    incr {
      set time [clock seconds]
      set data [getuser $user XTRA $configData(${what}field)]
      set olddata $data
      incr data $amount
      if {($data >= 0) && ($data <= $configData(${what}limit))} then {
        setuser $user XTRA [string toupper $configData(${what}field)] $data
        setuser $user XTRA [string toupper $configData(${what}itimefield)] $time
        setuser $user XTRA [string toupper $configData(${what}iinfofield)] "Last $uwhat increase: [clock format $time -format "%H:%M"]"
        return $data
      }
      return $olddata
    }
    decr {
      set time [clock seconds]
      set data [getuser $user XTRA $configData(${what}field)]
      set olddata $data
      incr data -$amount
      if {($data >= 0) && ($data <= $configData(${what}limit))} then {
        setuser $user XTRA [string toupper $configData(${what}field)] $data
        setuser $user XTRA [string toupper $configData(${what}dtimefield)] $time
        setuser $user XTRA [string toupper $configData(${what}dinfofield)] "Last $uwhat decrease: [clock format $time -format "%H:%M"]"
        return $data
      }
      return $olddata
    }
    set {
      if {($amount >= 0) && ($amount <= $configData(${what}limit))} then {
        setuser $user XTRA [string toupper $configData(${what}field)] $amount
        return $amount
      }
      return [getuser $user XTRA $configData(${what}field)]
    }
    list {
      return [getuser $user XTRA $configData(${what}field)]
    }
    itime {
      return [getuser $user XTRA $configData(${what}itimefield)]
    }
    dtime {
      return [getuser $user XTRA $configData(${what}dtimefield)]
    }
    check {
      set time [clock seconds]
      if {(![string length \
             [set data [getuser $user XTRA $configData(${what}field)]]]) || \
          ($data > $configData(${what}limit))} then {
        set data $configData(${what}limit)
        setuser $user XTRA [string toupper $configData(${what}field)] $data
      }
      if {(![string length \
             [set itime \
              [getuser $user XTRA $configData(${what}itimefield)]]]) || \
          ($itime > $time)} then {
        setuser $user XTRA [string toupper $configData(${what}itimefield)] $time
        set ireset 1
      }
      if {(![string length \
             [getuser $user XTRA $configData(${what}iinfofield)]]) || \
          ([info exists ireset])} then {
        setuser $user XTRA [string toupper $configData(${what}iinfofield)] "Last $uwhat increase: never"
      }
      if {(![string length \
             [set dtime \
              [getuser $user XTRA $configData(${what}dtimefield)]]]) || \
          ($dtime > $time)} then {
        setuser $user XTRA [string toupper $configData(${what}dtimefield)] $time
        set dreset 1
      }
      if {(![string length \
             [getuser $user XTRA $configData(${what}dinfofield)]]) || \
          ([info exists dreset])} then {
        setuser $user XTRA [string toupper $configData(${what}dinfofield)] "Last $uwhat decrease: never"
      }
      return $data
    }
    update {
      if {[isNumber \
           [set itime \
            [getuser $user XTRA $configData(${what}itimefield)]]]} then {
        setuser $user XTRA [string toupper $configData(${what}iinfofield)] [convertTime $itime]
      } else {
        setuser $user XTRA [string toupper $configData(${what}iinfofield)] "Last $uwhat increase: never"
      }
      if {[isNumber \
           [set dtime \
            [getuser $user XTRA $configData(${what}dtimefield)]]]} then {
        setuser $user XTRA [string toupper $configData(${what}dinfofield)] [convertTime $dtime]
      } else {
        setuser $user XTRA [string toupper $configData(${what}dinfofield)] "Last $uwhat decrease: never"
      }
      return 0
    }
    test1 {
      if {[getuser $user XTRA $configData(${what}field)] <= $configData(${what}ok)} then {
        return 0
      }
      return 1
    }
    test2 {
      if {[getuser $user XTRA $configData(${what}field)] >= $configData(${what}limit)} then {
        return 0
      }
      return 1
    }
    limit {
      return $configData(${what}limit)
    }
    clear {
      clearFields $user "$configData(${what}field) $configData(${what}itimefield) $configData(${what}iinfofield) $configData(${what}dtimefield) $configData(${what}dinfofield)"
      return 0
    }
  }
  return -1
}

##
## Update hp anf fp counts
##
## Args:    none
## Returns: nothing
##
proc updatePointData {} {
  variable configData

  set time [clock seconds]
  if {[string compare "" \
       [set users [userStatusData "" listusers status pack]]]} then {
    foreach user $users {
      if {![isNumber [set hp [userPointData $user list hp]]]} then {
        set hp [userPointData $user check hp]
      }
      if {![isNumber [set fp [userPointData $user list fp]]]} then {
        set fp [userPointData $user check fp]
      }
      if {(!$fp) && ($hp > 0)} then {
        if {![info exists hpdseconds]} then {
          set hpdseconds [convertSeconds from hours $configData(hpdecrtime)]
        }
        if {[expr $time - [userPointData $user dtime hp]] >= $hpdseconds} then {
          userPointData $user decr hp 1
        }
      } elseif {$fp > 0} then {
        if {![info exists fpdseconds]} then {
          set fpdseconds [convertSeconds from hours $configData(fpdecrtime)]
        }
        if {[expr $time - [userPointData $user dtime fp]] >= $fpdseconds} then {
          userPointData $user decr fp 1
        }
      }
      if {($hp >= 0) && ($hp < $configData(hplimit)) && \
          ($fp >= $configData(fpok))} then {
        if {![info exists hpiseconds]} then {
          set hpiseconds [convertSeconds from hours $configData(hpincrtime)]
        }
        if {[expr $time - [userPointData $user itime hp]] >= $hpiseconds} then {
          userPointData $user incr hp 1
        }
      }
    }
  }
  return
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      fpdinfofield -
      fpdtimefield -
      fpfield -
      fpiinfofield -
      fpitimefield -
      hpdinfofield -
      hpdtimefield -
      hpfield -
      hpiinfofield -
      hpitimefield {
        moveFields $oldConfigData($name2) [string toupper $configData($name2)]
        whoisFields update $configData($name2) $oldConfigData($name2)
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  whoisFields update "$configData(hpfield) $configData(hpiinfofield) $configData(hpdinfofield) $configData(fpfield) $configData(fpiinfofield) $configData(fpdinfofield)" \
                     "$oldConfigData(hpfield) $oldConfigData(hpiinfofield) $oldConfigData(hpdinfofield) $oldConfigData(fpfield) $oldConfigData(fpiinfofield) $oldConfigData(fpdinfofield)"
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  whoisFields remove "$configData(hpfield) $configData(hpiinfofield) $configData(hpdinfofield) $configData(fpfield) $configData(fpiinfofield) $configData(fpdinfofield)"
  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  fpdecrtime {integer 24 {} hours {time in hours between fp drops for a user}}
  fpdinfofield {word fpdinfo {} field {field for user's last fp decrease info}}
  fpdtimefield {word fpdtime {} field {field for user's last fp decrease time}}
  fpfield {word fp {} field {field for user's fp}}
  fpiinfofield {word fpiinfo {} field {field for user's last fp increase info}}
  fpitimefield {word fpitime {} field {field for user's last fp increase time}}
  fplimit {integer 10 {} limit {max number of foodpoints a user has}}
  fpok {integer 5 {} amount {amount that a user's fp must reach before it registers as ok}}
  hpdecrtime {integer 24 {} hours {time in hours between hp drops for a user if fp is 0}}
  hpdinfofield {word hpdinfo {} field {field for user's last hp decrease info}}
  hpdtimefield {word hpdtime {} field {field for user's last hp decrease time}}
  hpfield {word hp {} field {field for user's hp}}
  hpiinfofield {word hpiinfo {} field {field for user's last hp increase info}}
  hpincrtime {integer 48 {} hours {hours before a user's hp is increased}}
  hpitimefield {word hpitime {} field {field for user's last hp increase time}}
  hplimit {integer 10 {} limit {max amount of hitpoints a user has}}
  hpok {integer 5 {} amount {amount that a user's hp must reach before it registers as ok}}
}
