###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2005  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: statusdata.tcl,v 1.13 2005/01/04 03:36:44 tothwolf Exp $
##

# name: wrpg-statusdata
# version: 1.0
# author: Tothwolf <tothwolf@concentric.net>
# license: GNU GPL v2
# config: wrpg-statusdata.conf
# description: Wolf RPG user status data support.
# requires: clearFields findField isNumber convertSeconds whoisFields wpLog
# provides: userStatusData userAssessmentData updateStatusData


##
## Handle a user's status/level/position data
##
## Args:    user, action {set|clear|compare|list|listusers|check}
##          what {status|level|pos}, data
## Returns:
##
proc userStatusData {user action what args} {
  variable configData

  switch -exact $action {
    set {
      if {[string compare [string tolower [set data [lindex $args 0]]] \
           [string tolower [getuser $user XTRA $configData(${what}field)]]]} then {
        setuser $user XTRA [string toupper $configData(${what}field)] $data
        return 1
      }
    }
    clear {
      clearFields $user $configData(${what}field)
      return 1
    }
    compare {
      if {([string length [set data \
            [getuser $user XTRA $configData(${what}field)]]]) && \
           (![string compare [string tolower $data] \
              [string tolower [lindex $args 0]]])} then {
        return 1
      }
    }
    list {
      return [getuser $user XTRA $configData(${what}field)]
    }
    listusers {
      set list ""
      if {[string compare "" [set data [string tolower [lindex $args 0]]]]} then {
        foreach user [lsort [findField $configData(${what}field)]] {
          if {![string compare $data \
                [string tolower \
                 [getuser $user XTRA $configData(${what}field)]]]} then {
            lappend list $user
          }
        }
      } else {
        foreach user [lsort [findField $configData(${what}field)]] {
          lappend list $user
        }
      }
      return $list
    }
    check {
      if {[info exists configData(default${what})]} then {
        setuser $user XTRA [string toupper $configData(${what}field)] $configData(default${what})
      }
    }
  }
  return 0
}

##
## Handle a user's assessment data
##
## Args:    user, action {add|del|started|time|timeleft|update}, data
## Returns:
##
proc userAssessmentData {user action args} {
  variable configData

  switch -exact -- $action {
    add {
      set time [clock seconds]
      set seconds [convertSeconds from days $configData(asdays)]
      setuser $user XTRA $configData(astimefield) $time
      if {$configData(asdays)} then {
        set info "Assessment ends: [clock format [expr $time + $seconds] -format "%a %b %d %Y"], Time left: [duration $seconds]"
      } else {
        set info "Assessment started: [clock format $time -format "%a %b %d %Y"]"
      }
      setuser $user XTRA [string toupper $configData(asinfofield)] $info
      return 1
    }
    del {
      clearFields $user "$configData(astimefield) $configData(asinfofield)"
      return 1
    }
    started {
      if {[isNumber [set time [getuser $user XTRA $configData(astimefield)]]]} then {
        return [clock format $time -format "%a %b %d %Y"]
      }
      return
    }
    time {
      if {[isNumber [set time [getuser $user XTRA $configData(astimefield)]]]} then {
        return $time
      }
      return 0
    }
    timeleft {
      if {$configData(asdays)} then {
        set astime [getuser $user XTRA $configData(astimefield)]
        set time [clock seconds]
        set timeleft [expr ($astime + [convertSeconds from days $configData(asdays)]) - $time]
        if {$timeleft > 0} then {
          return $timeleft
        } else {
          return 0
        }
      } else {
        return -1
      }
    }
    update {
      if {$configData(asdays)} then {
        set astime [getuser $user XTRA $configData(astimefield)]
        set time [clock seconds]
        set endtime [expr ($astime + [convertSeconds from days $configData(asdays)])]
        set timeleft [expr $endtime - $time]
        if {$timeleft > 0} then {
          set info "Assessment ends: [clock format $endtime -format "%a %b %d %Y"], time left: [duration $timeleft]"
        } else {
          set info "Assessment ended: [clock format $endtime -format "%a %b %d %Y"]"
        }
        setuser $user XTRA [string toupper $configData(asinfofield)] $info
      }
      return 1
    }
  }
  return
}

##
## Check/update assessment time and status
##
## Args:    none
## Returns: nothing
##
proc updateStatusData {} {
  variable configData

  if {($configData(asdays)) && \
      ([string compare "" \
        [set users [userStatusData "" listusers status assessment]]])} then {
    foreach user $users {
      if {($configData(autoupdate)) && \
          (![userAssessmentData $user timeleft])} then {
        if {[userStatusData $user set status pack]} then {
          userAssessmentData $user del
          userStatusData $user check pos
          userStatusData $user check level
          wpLog o * "Updated status for $user from assessment to pack."
        } else {
          wpLog o * "Error updating status for $user from assessment to pack."
        }
      } else {
        userAssessmentData $user update
      }
    }
  }
  return
}

##
## Trace configuration changes
##
## Args:    name1, name2, operation
## Returns: nothing
##
proc TraceConfigData {name1 name2 op} {
  variable configData
  variable oldConfigData

  if {[string compare $oldConfigData($name2) $configData($name2)]} then {
    switch -exact -- $name2 {
      statusfield -
      levelfield -
      posfield -
      asinfofield {
        moveFields $oldConfigData($name2) [string toupper $configData($name2)]
        whoisFields update $configData($name2) $oldConfigData($name2)
      }
    }
    set oldConfigData($name2) $configData($name2)
  }
  return
}

##
## Init module specific things at load time
##
## Args:    none
## Returns: nothing
##
proc ModuleInit {} {
  variable configData
  variable oldConfigData

  foreach {name data} [array get configData] {
    if {![info exists oldConfigData($name)]} then {
      set oldConfigData($name) $data
    }
  }
  whoisFields update "$configData(statusfield) $configData(levelfield) $configData(posfield) $configData(asinfofield)" \
                     "$oldConfigData(statusfield) $oldConfigData(levelfield) $oldConfigData(posfield) $oldConfigData(asinfofield)"
  trace variable configData w [namespace code TraceConfigData]
  return
}

##
## Destroy module specific things before unload
##
## Args:    none
## Returns: nothing
##
proc ModuleDestroy {} {
  variable configData

  whoisFields remove "$configData(statusfield) $configData(levelfield) $configData(posfield) $configData(asinfofield)"
  trace vdelete configData w [namespace code TraceConfigData]
  return
}

# Configuration defaults, values and descriptions
array set configDefaults {
  autoupdate {switch 1 {} {0/1} {automatically update a user's status from assessment to pack?}}
  asdays {integer 30 {} days {how many days assessments will last (0 for open ended)}}
  asinfofield {word wrpgasinfo {} field {field to store assessment info}}
  astimefield {word wrpgastime {} field {field to store assessment time}}
  defaultlevel {integer 1 {} level {default level for user}}
  defaultpos {string {not assigned} {} position {default position for user}}
  levelfield {word wrpglevel {} field {field to store status level}}
  posfield {word wrpgpos {} field {field to store position info}}
  statusfield {word wrpgstatus {} field {field to store user status}}
}
